<?php

/**
 * @am_plugin_api 6.0
 * https://www.paypalobjects.com/.well-known/openid-configuration
 */
class Am_Plugin_PaypalIdentity extends Am_Plugin
{
    const PLUGIN_STATUS = self::STATUS_BETA;
    const PLUGIN_COMM = self::COMM_COMMERCIAL;
    const PLUGIN_REVISION = '6.3.21';

    const LOGIN_BUTTON = 'https://www.paypalobjects.com/webstatic/en_US/developer/docs/lipp/loginwithpaypalbutton.png';
    const LIVE_ENDPOINT = 'https://www.paypal.com/';
    const SANDBOX_ENDPOINT = 'https://www.sandbox.paypal.com/';
    const LIVE_API_ENDPOINT = 'https://api.paypal.com/';
    const SANDBOX_API_ENDPOINT = 'https://api.sandbox.paypal.com/';

    const DATA_KEY = 'paypal-identity-user_id';

    protected $_isDebug = false;
    protected $_configPrefix = 'misc.';

    function _initSetupForm(Am_Form_Setup $form)
    {
        $form->setTitle('PayPal Identity');

        $form->addText('client_id', ['class' => 'am-el-wide'])
            ->setLabel('Client ID');
        $form->addSecretText('secret', ['class' => 'am-el-wide'])
            ->setLabel('Secret');
        $form->addAdvCheckbox('seamless_checkout')
            ->setLabel("Enable Seamless Checkout\n" .
                "this option works with PayPal Express payment plugin.\n" .
                "Your PayPal application should have option Use Seamless Checkout enabled");
        $form->addAdvCheckbox("testing")
            ->setLabel("Sandbox", "Use Sandbox Environment");

        $form->addSelect("login_postion")
            ->setLabel("Login Button Position")
            ->loadOptions([
                'login/form/before' => ___("Above Login Form"),
                'login/form/after' => ___("Below Login Form")
            ]);

        $form->addMagicSelect('add_access', ['class' => 'am-combobox-fixed'])
            ->setLabel("Add Access\nto user who sign up via PayPal")
            ->loadOptions($this->getDi()->productTable->getOptions());
    }

    function isConfigured()
    {
        return $this->getConfig('client_id') && $this->getConfig('secret');
    }

    function getEndpoint()
    {
        return $this->getConfig('testing') ?
            self::SANDBOX_ENDPOINT :
            self::LIVE_ENDPOINT;
    }

    function getApiEndpoint()
    {
        return $this->getConfig('testing') ?
            self::SANDBOX_API_ENDPOINT :
            self::LIVE_API_ENDPOINT;
    }

    function getLoginUrl($b)
    {
        $state = $this->getDi()->security->randomString(30);

        $ns = $this->getDi()->session->ns($this->getId());
        $ns->{$state} = $b;

        return $this->getEndpoint() . 'signin/authorize?' . http_build_query([
            'client_id' => $this->getConfig('client_id'),
            'response_type' => 'code',
            'scope' => 'email' . ($this->getConfig('seamless_checkout') ? ' https://uri.paypal.com/services/expresscheckout' : ''),
            'redirect_uri' => $this->getDi()->surl("misc/{$this->getId()}/callback", false),
            'nonce' => time() . base64_encode($this->getDi()->security->randomString(15)),
            'state' => $state
            ]);
    }

    function onInitFinished(Am_Event $event)
    {
        if (!$this->getDi()->auth->getUserId()) {
            $this->getDi()->blocks->add(
                $this->getConfig('login_postion', 'login/form/before'),
                new Am_Block_Base(null, 'misc-paypal-identity-signin', $this, [$this, 'renderLoginButton'])
            );
            $this->getDi()->blocks->add(
                'signup/form/before',
                new Am_Block_Base(null, 'misc-paypal-identity-signup', $this, [$this, 'renderLoginButton'])
            );
        }
    }

    function renderLoginButton(Am_View $v)
    {
        $url = $this->getDi()->url("misc/{$this->getId()}/redirect", [
            'b' => $_SERVER['REQUEST_URI'],
            'h' => $this->getDi()->security->siteHash($_SERVER['REQUEST_URI'], 5)
        ]);
        return sprintf(<<<CUT
<div class="am-paypal-identity-button-wrapper">
    <div class="am-paypal-identity-button">
        <a href="%s"><img src="%s" /></a>
    </div>
</div>
CUT
            ,Am_Html::escape($url), Am_Html::escape(self::LOGIN_BUTTON));
    }

    function directAction($request, $response, $invokeArgs)
    {
        switch($request->getActionName()) {
            case 'callback':
                return $this->callbackAction($request);
            case 'redirect':
                return $this->redirectAction($request);
        }

        throw new Am_Exception_InputError;
    }

    function callbackAction($request)
    {
        $this->log([$request], 'REQUEST');

        $code = $request->getParam('code');
        $state = $request->getParam('state');

        $ns = $this->getDi()->session->ns($this->getId());
        if (!isset($ns->{$state})) {
            throw new Am_Exception_InputError;
        }
        $b = $ns->{$state};
        unset($ns->{$state});

        $req = new Am_HttpRequest($this->getApiEndpoint() . 'v1/identity/openidconnect/tokenservice', Am_HttpRequest::METHOD_POST);
        $req->setAuth($this->getConfig('client_id'), $this->getConfig('secret'));
        $req->addPostParameter([
            'grant_type' => 'authorization_code',
            'code' => $code,
            'redirect_uri' => $this->getDi()->surl("misc/{$this->getId()}/callback", false),
        ]);
        $resp = $req->send();

        $this->log([$req, $resp], 'ACCESS TOKEN');

        if ($resp->getStatus() == '200' && ($payload = json_decode($resp->getBody(), true))) {
            $access_token = $payload['access_token'];
            $req = new Am_HttpRequest($this->getApiEndpoint() . 'v1/identity/openidconnect/userinfo?schema=openid');
            $req->setHeader('Authorization', "Bearer $access_token");

            $resp = $req->send();
            $this->log([$req, $resp], 'PROFILE');
            if ($resp->getStatus() == '200' && ($payload = json_decode($resp->getBody(), true))) {

                $is_created = false;

                do {
                    if ($user = $this->getDi()->userTable->findFirstByData(self::DATA_KEY, $payload['user_id'])) {
                        break;
                    }

                    if ($user = $this->getDi()->userTable->findFirstByEmail($payload['email'])) {
                        break;
                    }

                    $user = $this->getDi()->userRecord;
                    $user->email = $payload['email'];
                    $user->generateLogin();
                    $user->generatePassword();
                    $is_created = true;
                    break;
                } while (true);

                $user->data()->set(self::DATA_KEY, $payload['user_id']);
                $user->save();

                if ($is_created) {
                    if ($pids = $this->getConfig('add_access')) {
                        foreach ($pids as $pid) {
                            if ($product = $this->getDi()->productTable->load($pid, false)) {
                                $billingPlan = $product->getBillingPlan();

                                $access = $this->getDi()->accessRecord;
                                $access->product_id = $pid;
                                $access->begin_date = $this->getDi()->sqlDate;

                                $period = new Am_Period($billingPlan->first_period);
                                $access->expire_date = $period->addTo($access->begin_date);

                                $access->user_id = $user->pk();
                                $access->comment = 'from PayPal login';
                                $access->insert();
                            }
                        }
                        $user->checkSubscriptions();
                    }
                }

                if ($this->getConfig('seamless_checkout')) {
                    //access token can be used only during first 15 minutes
                    $this->getDi()->store->set('paypal-access-token-' . $user->pk(), $access_token, sqlTime('+14 minutes'));
                }
                $adapter = new Am_Auth_Adapter_User($user);
                $this->getDi()->auth->login($adapter, $request->getClientIp(), false);
                Am_Mvc_Response::redirectLocation($b);
            }
        } else {
            throw new Am_Exception_InputError;
        }
    }

    function redirectAction($request)
    {
        $b = $request->getParam('b');
        $h = $request->getParam('h');
        if ($this->getDi()->security->siteHash($b, 5) != $h) {
            throw new Am_Exception_InputError;
        }

        Am_Mvc_Response::redirectLocation($this->getLoginUrl($b));
    }

    function log($events, $title)
    {
        if (!$this->_isDebug)
            return;
        $l = $this->getDi()->invoiceLogRecord;
        $l->paysys_id = $this->getId();
        $l->title = $title;
        foreach ($events as $event) {
            $l->add($event);
        }
    }
}