<?php
/**
 * @table integration
 * @id mybb
 * @title MyBB
 * @visible_link http://www.mybb.com/
 * @description MyBB is a discussion board that has been around for a while;
 * it has evolved from other bulletin boards into the forum package it is today.
 * Therefore, it is a professional and efficient discussion board, developed
 * by an active team of developers.
 * @different_groups 1
 * @single_login 1
 * @type Bulletin Boards
 * @am_protect_api 6.0
 */
class Am_Protect_Mybb extends Am_Protect_Databased
{
    const PLUGIN_STATUS = self::STATUS_BETA;
    const PLUGIN_COMM = self::COMM_COMMERCIAL; //paid
    const PLUGIN_REVISION = '6.3.21';

    protected $groupMode = self::GROUP_MULTI;
    protected $guessTablePattern = "users";
    protected $guessFieldsPattern = [
        'uid',
        'username',
        'avatardimensions',
        'usernotes'
    ];

    const MYBB = 'mybb';

    function parseExternalConfig($path)
    {
        $config_path = $path . "/inc/config.php";
        $config=null;
        if (!is_file($config_path) || !is_readable($config_path))
        {
            throw new Am_Exception_InputError("Specified path is not a valid MyBB installation!");
        }
        include($config_path);
        if (!is_array($config))
        {
            throw new Am_Exception_InputError("Specified path is not a valid MyBB installation!");
        }
        return [
            'db' => $config['database']['database'],
            'host' => $config['database']['hostname'],
            'user' => $config['database']['username'],
            'pass' => $config['database']['password'],
            'prefix' => $config['database']['table_prefix']
        ];
    }

    public function getPasswordFormat()
    {
        return self::MYBB;
    }

    public function getAvailableUserGroupsSql()
    {
        return "SELECT 
            gid as id,
            title as title, 
            (gid=7) as is_banned,
            (gid IN (3,4,6)) as is_admin
            FROM ?_usergroups";
    }

    function loginKey()
    {
        return $this->randomStr(50);
    }

    function randomStr($length="8")
    {
        $set = ["a", "A", "b", "B", "c", "C", "d", "D", "e", "E", "f", "F", "g", "G", "h", "H", "i", "I", "j", "J", "k", "K", "l", "L", "m", "M", "n", "N", "o", "O", "p", "P", "q", "Q", "r", "R", "s", "S", "t", "T", "u", "U", "v", "V", "w", "W", "x", "X", "y", "Y", "z", "Z", "1", "2", "3", "4", "5", "6", "7", "8", "9"];
        $str = '';

        for ($i = 1; $i <= $length; ++$i)
        {
            $ch = rand(0, count($set) - 1);
            $str .= $set[$ch];
        }

        return $str;
    }

    public function ipLong(User $user)
    {
        return ip2long($user->remote_addr);
    }

    public function createTable()
    {
        $table = new Am_Protect_Table_Mybb($this, $this->getDb(), '?_users', 'uid');
        $table->setFieldsMapping([
            [Am_Protect_Table::FIELD_LOGIN, 'username'],
            [Am_Protect_Table::FIELD_PASS, 'password'],
            [Am_Protect_Table::FIELD_SALT, 'salt'],
            [[$this, 'loginKey'], 'loginkey'],
            [Am_Protect_Table::FIELD_EMAIL, 'email'],
            [':' . $this->getDi()->time, 'regdate'],
            [':' . $this->getDi()->time, 'lastactive'],
            [':' . $this->getDi()->time, 'lastvisit'],
            [':1', 'allownotices'],
            [':1', 'hideemail'],
            [':1', 'receivepms'],
            [':1', 'pmnotice'],
            [':1', 'pmnotify'],
            [':1', 'showimages'],
            [':1', 'showvideos'],
            [':1', 'showsigs'],
            [':1', 'showavatars'],
            [':1', 'showquickreply'],
            [':1', 'showredirect'],
            [':2', 'subscriptionmethod'],
            [Am_Protect_Table::FIELD_REMOTE_ADDR, 'regip']
        ]);


        return $table;
    }

    public function createSessionTable()
    {
        $table = new Am_Protect_SessionTable_Mybb($this, $this->getDb(), '?_sessions', 'sid');
        $table->setTableConfig($a = [
            Am_Protect_SessionTable::FIELD_SID => 'sid',
            Am_Protect_SessionTable::FIELD_UID => 'uid',
            Am_Protect_SessionTable::FIELD_IP => 'ip',
            Am_Protect_SessionTable::FIELD_CHANGED => 'time',
            Am_Protect_SessionTable::FIELD_UA => 'useragent',
            Am_Protect_SessionTable::SESSION_COOKIE => 'sid',
            Am_Protect_SessionTable::COOKIES_ADDITIONAL => [
                'mybbuser' => function($userRecord) {return $userRecord->uid."_".$userRecord->loginkey;}
            ],
            Am_Protect_SessionTable::COOKIE_PARAMS => [
                'domain' => $_SERVER['HTTP_HOST']
            ]
        ]);
        return $table;
    }

    function cryptPassword($pass, &$salt = null, User $user = null)
    {
        if (!$salt)
            $salt = $this->randomStr(8);
        return md5(md5($salt) . md5($pass));
    }


}

class Am_Protect_Table_Mybb extends Am_Protect_Table
{

    public function updateStats(Am_Record $record)
    {
        $stats = $this->_db->selectRow("select * from ?_datacache where title = 'stats'");
        $stats['cache'] = unserialize($stats['cache']);
        $stats['cache']['numusers'] = $this->_db->selectCell("select count(*) from ?_users");
        $stats['cache']['lastuid'] = $record->uid;
        $stats['cache']['lastusername'] = $record->username;
        $this->_db->query("update ?_datacache set cache=? where title='stats'", serialize($stats['cache']));
    }

    public function insertFromAmember(User $user, SavedPass $pass, $groups)
    {
        $record = parent::insertFromAmember($user, $pass, $groups);
        $this->updateStats($record);
    }

    public function getGroups(Am_Record $record)
    {
        $groups = [$record->usergroup];
        $groups = array_merge($groups, explode(',', $record->additionalgroups));
        return array_filter($groups);
    }

    public function setGroups(Am_Record $record, $groups)
    {
        $record->usergroup = $record->displaygroup = array_shift($groups);
        $record->additionalgroups = implode(',', $groups);
        $record->updateSelectedFields(['usergroup', 'additionalgroups', 'displaygroup']);
    }

}
class Am_Protect_SessionTable_Mybb extends Am_Protect_SessionTable
{
    function getUserAgent()
    {
        return substr($this->getDi()->request->getServer('HTTP_USER_AGENT'),0,100);
    }
}