<?php

/**
 * @table integration
 * @id xenforo
 * @title XenForo
 * @description Popular Bulletin Board Software
 * @type Bulletin Boards
 * @single_login 1
 * @different_groups 1
 * @hidden_link http://xenforo.com/
 * @visible_link http://xenforo.com/
 *
 * @am_protect_api 6.0
 */
class Am_Protect_Xenforo extends Am_Protect_Databased
{
    const PLUGIN_STATUS = self::STATUS_PRODUCTION;
    const PLUGIN_REVISION = '6.3.21';
    const DEFAULT_VERSION = 104;

    /** @var string table name without prefix */
    protected $guessTablePattern = 'user';

    /** @var array of several fieldnames in the table */
    protected $guessFieldsPattern = [
        'username', 'custom_title'
    ];
    protected $groupMode = self::GROUP_MULTI;
    protected $_options = null;

    public function init()
    {
        $this->getDi()->userTable->customFields()->add(
            new Am_CustomFieldCheckbox('xenforo_admin', 'Xenforo Admin?'));
        parent::init();
    }

    function getPasswordFormat()
    {
        switch ($this->getConfig('version', self::DEFAULT_VERSION)) {
            case 200:
                return SavedPassTable::PASSWORD_PASSWORD_HASH;
            case 130:
                return SavedPassTable::PASSWORD_PHPASS;
            case 104:
            default:
                return SavedPassTable::PASSWORD_SHA256_SHA256_PASS_SALT;
        }
    }

    public function afterAddConfigItems($form)
    {
        parent::afterAddConfigItems($form);
        $form->addText("protect.{$this->getId()}.cookie_prefix")->setLabel("Cookie Prefix\n" .
            "do not change default value unless you know what you do");
        $form->setDefault("protect.{$this->getId()}.cookie_prefix", 'xf_');
        $form->addSelect("protect.{$this->getId()}.version", [], ["options" => ['104' => '1.0.4', '130' => '1.3.0', '200' => '2 or above']])
            ->setLabel("Xenforo version");
        $form->addText("protect.{$this->getId()}.global_salt", ['class'=>'am-el-wide'])->setLabel("Global Salt\n" .
            "can be found in your xenforo installatioin in file\n" .
            "Version 1: /library/XenForo/Application.php\n" .
            "Version 2: /src/XF/App.php");
    }

    public function parseExternalConfig($path)
    {
        if (!is_file($config_path = $path . "/library/config.php") || !is_readable($config_path))
        {
            if (!is_file($config_path = $path . "/src/config.php") || !is_readable($config_path))
                throw new Am_Exception_InputError("This is not a valid " . $this->getTitle() . " installation");
            include_once $config_path;

            $content = file_get_contents($path . "/src/XF/App.php");
            preg_match("/'globalSalt'\s*=>\s*'(.+?)'/im", $content, $m);

            return [
                'version' => 200,
                'db' => $config['db']['dbname'],
                'prefix' => 'xf_',
                'host' => $config['db']['host'],
                'user' => $config['db']['username'],
                'pass' => $config['db']['password'],
                'global_salt' => $m[1]
            ];

        }

        include_once $config_path;

        $content = file_get_contents($path . "/library/XenForo/Application.php");
        preg_match("/'globalSalt'\s*=>\s*'(.+?)'/im", $content, $m);

        return [
            'db' => $config['db']['dbname'],
            'prefix' => 'xf_',
            'host' => $config['db']['host'],
            'user' => $config['db']['username'],
            'pass' => $config['db']['password'],
            'global_salt' => $m[1]
        ];
    }

    public function canAutoCreate()
    {
        return true;
    }

    public function getLoggedInRecord()
    {
        $session_id = @$_COOKIE[$this->getSessionCookieName()];
        if (!$session_id)
            return null;

        $sess = $this->getDb()->selectRow("SELECT * FROM ?_session WHERE session_id=?", $session_id);
        if (!$sess || $sess['expiry_date'] < $this->getDi()->time)
            return null;

        $data = unserialize($sess['session_data']);

        if (isset($data['user_id']) && $data['user_id']) {
            $row = $this->getDb()->selectRow("SELECT * FROM ?_user WHERE user_id=?", $data['user_id']);
            if ($row) {
                return $this->getTable()->createRecord($row);
            }
        }
        elseif (isset($data['userId']) && $data['userId']) {
            $row = $this->getDb()->selectRow("SELECT * FROM ?_user WHERE user_id=?", $data['userId']);
            if ($row) {
                return $this->getTable()->createRecord($row);
            }
        }
    }

    function loginUser(Am_Record $record, $password)
    {
        $cookie_name = $this->getSessionCookieName();

        $session_id = $_COOKIE[$cookie_name] ?? md5(uniqid(microtime(true), true));
        $sess = $this->getDb()->selectRow("SELECT * FROM ?_session WHERE session_id=?", $session_id);
        $isExist = true;
        if (!$sess) {
            $isExist = false;
            $sess = [
                'session_id' => $session_id,
                'session_data' => $this->getDefaultSessionData(),
                'expiry_date' => $this->getDi()->time + 24 * 3600
            ];
            Am_Cookie::set($cookie_name, $session_id, 0, '/', $_SERVER['HTTP_HOST']);
        }
        else {
            $sess['session_data'] = unserialize($sess['session_data']);
        }

        $ip = '';
        switch ($this->getConfig('version', self::DEFAULT_VERSION)) {
            case 104:
                $ip = ip2long($_SERVER['REMOTE_ADDR']);
                break;
            case 200:
            case 130:
                $ip = $this->convertIpStringToBinary($_SERVER['REMOTE_ADDR']);
                break;
        }

        if($this->getConfig('version', self::DEFAULT_VERSION) == 200)
        {
            $sess['session_data']['userId'] = $record->user_id;
            $sess['session_data']['_ip'] = $ip;
            if($passwordDate = $this->getDb()->selectCell("SELECT password_date FROM ?_user_profile WHERE user_id=?", $record->user_id))
                $sess['session_data']['passwordDate'] = $passwordDate;
        }
        else
        {
            $sess['session_data']['user_id'] = $record->user_id;
            $sess['session_data']['ip'] = $ip;
        }

        $sess['session_data'] = serialize($sess['session_data']);

        $isExist ?
                $this->getDb()->query("UPDATE ?_session SET session_data=?, expiry_date=? WHERE session_id=?", $sess['session_data'], $this->getDi()->time + 24 * 3600, $sess['session_id']) :
                $this->getTable()->bulkInsert('?_session', $sess);

        if ($this->getDi()->config->get('protect.php_include.remember_auto', false)) {
            if ($this->getConfig('version', self::DEFAULT_VERSION) == 200) {
                do {
                    $key = Am_Di::getInstance()->security->randomString(40);
                    $dbKey = hash('sha256', $key, true);

                    $exists = $this->getDb()->selectCell("SELECT count(*) from ?_user_remember WHERE user_id= ? AND remember_key = ?", $record->user_id, $dbKey);
                } while ($exists);
                $this->getDb()->query("INSERT INTO ?_user_remember (user_id, remember_key, start_date, expiry_date) VALUES(?, ?, ?, ?)"
                        , $record->user_id
                        , $dbKey
                        , Am_Di::getInstance()->time
                        , Am_Di::getInstance()->time + 28 * 86400);
                $remember = $record->pk() . ',' . $key;
            } else {
                $auth = $this->getDb()->selectRow('SELECT * FROM ?_user_authenticate WHERE user_id=?', $record->pk());
                $remember = $record->pk() . ',' . sha1($this->getConfig('global_salt') . $auth['remember_key']);
            }
            Am_Cookie::set($this->getConfig('cookie_prefix') . 'user', $remember, $this->getDi()->time + $this->getDi()->config->get('protect.php_include.remember_period', 60) * 3600 * 24, '/', $_SERVER['HTTP_HOST']);
        }

        return true;
    }

    function logoutUser(User $user)
    {
        $cookie_name = $this->getSessionCookieName();
        if (isset($_COOKIE[$cookie_name])) {
            $this->getDb()->query("DELETE FROM ?_session WHERE session_id=?", $_COOKIE[$cookie_name]);
        }
        Am_Cookie::delete($cookie_name);
        Am_Cookie::set($this->getConfig('cookie_prefix') . 'user', null, time() - 24 * 3600, '/', $_SERVER['HTTP_HOST']);
    }

    function getSessionCookieName()
    {
        return $this->getConfig('cookie_prefix') . 'session';
    }

    public function convertIpStringToBinary($ip)
    {
        $originalIp = $ip;
        $ip = trim($ip);

        if (strpos($ip, ':') !== false) {
            // IPv6
            if (preg_match('#:(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})$#', $ip, $match)) {
                // embedded IPv4
                $long = ip2long($match[1]);
                if (!$long) {
                    return false;
                }

                $hex = str_pad(dechex($long), 8, '0', STR_PAD_LEFT);
                $v4chunks = str_split($hex, 4);
                $ip = str_replace($match[0], ":$v4chunks[0]:$v4chunks[1]", $ip);
            }

            if (strpos($ip, '::') !== false) {
                if (substr_count($ip, '::') > 1) {
                    // ambiguous
                    return false;
                }

                $delims = substr_count($ip, ':');
                if ($delims > 7) {
                    return false;
                }

                $ip = str_replace('::', str_repeat(':0', 8 - $delims) . ':', $ip);
                if ($ip[0] == ':') {
                    $ip = '0' . $ip;
                }
            }

            $ip = strtolower($ip);

            $parts = explode(':', $ip);
            if (count($parts) != 8) {
                return false;
            }

            foreach ($parts AS &$part) {
                $len = strlen($part);
                if ($len > 4 || preg_match('/[^0-9a-f]/', $part)) {
                    return false;
                }

                if ($len < 4) {
                    $part = str_repeat('0', 4 - $len) . $part;
                }
            }

            $hex = implode('', $parts);
            if (strlen($hex) != 32) {
                return false;
            }

            return $this->convertHexToBin($hex);
        }
        else if (strpos($ip, '.')) {
            // IPv4
            if (!preg_match('#(\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})#', $ip, $match)) {
                return false;
            }

            $long = ip2long($match[1]);
            if (!$long) {
                return false;
            }

            return $this->convertHexToBin(
                    str_pad(dechex($long), 8, '0', STR_PAD_LEFT)
            );
        }
        else if (strlen($ip) == 4 || strlen($ip) == 16) {
            // already binary encoded
            return $ip;
        }
        else if (is_numeric($originalIp) && $originalIp < pow(2, 32)) {
            // IPv4 as integer
            return $this->convertHexToBin(
                    str_pad(dechex($originalIp), 8, '0', STR_PAD_LEFT)
            );
        }
        else {
            return false;
        }
    }

    public function convertHexToBin($hex)
    {
        if (function_exists('hex2bin')) {
            return hex2bin($hex);
        }

        $len = strlen($hex);

        if ($len % 2) {
            trigger_error('Hexadecimal input string must have an even length', E_USER_WARNING);
        }

        if (strspn($hex, '0123456789abcdefABCDEF') != $len) {
            trigger_error('Input string must be hexadecimal string', E_USER_WARNING);
        }

        return pack('H*', $hex);
    }

    function getDefaultSessionData()
    {
        switch ($this->getConfig('version', self::DEFAULT_VERSION)) {
            case 104:
                return [
                    'sessionStart' => $this->getDi()->time,
                    'ip' => ip2long($_SERVER['REMOTE_ADDR']),
                    'userAgent' => $_SERVER['HTTP_USER_AGENT'],
                    'referer' => $_SERVER['HTTP_REFERER']
                ];
                break;
            case 130:
                return [
                    'sessionStart' => $this->getDi()->time,
                    'ip' => $this->convertIpStringToBinary($_SERVER['REMOTE_ADDR']),
                    'sessionCsrf' => $this->getDi()->security->randomString(10)
                ];
                break;
            case 200:
                return [
                    '_ip' => $this->convertIpStringToBinary($_SERVER['REMOTE_ADDR']),
                ];
                break;
        }
    }

    public function getAvailableUserGroupsSql()
    {
        $admin_groups = array_filter(explode(',', $this->getConfig('admin_group')), 'intval');
        $admin_groups[] = -1;
        $admin_groups = implode(',', $admin_groups);
        $banned_groups = array_filter(explode(',', $this->getConfig('banned_group')), 'intval');
        $banned_groups[] = -1;
        $banned_groups = implode(',', $banned_groups);

        return "SELECT user_group_id as id, title as title,
                (user_group_id in (" . $admin_groups . ")) as is_admin,
                (user_group_id in (" . $banned_groups . ")) as is_banned
                from ?_user_group";
    }

    public function createTable()
    {
        $table = new Am_Protect_Table_Xenforo($this, $this->getDb(), '?_user', 'user_id');
        $table->setFieldsMapping([
            [Am_Protect_Table::FIELD_LOGIN, 'username'],
            [Am_Protect_Table::FIELD_EMAIL, 'email'],
            [Am_Protect_Table::FIELD_ADDED_STAMP, 'register_date'],
            [Am_Protect_Table::FIELD_ADDED_STAMP, 'last_activity'],
            [':0', 'language_id'],
            [':0', 'style_id'],
            [':Europe/London', 'timezone'],
            [':1', 'visible'],
            [':0', 'user_group_id'], //will be updated late by setGroups
            [':', 'secondary_group_ids'],
            [':0', 'display_style_group_id'],
            [':0', 'permission_combination_id']
        ]);
        $table->setGroupsTableConfig([
            Am_Protect_Table::GROUP_TABLE => '?_user_group_relation',
            Am_Protect_Table::GROUP_UID => 'user_id',
            Am_Protect_Table::GROUP_GID => 'user_group_id',
            Am_Protect_Table::GROUP_ADDITIONAL_FIELDS => [
                //we will update it to 1 for needed group later @see Am_Protect_Table_Xenforo::setGroups
                'is_primary' => 0
            ]
        ]);
        return $table;
    }

    public function getOption($name)
    {
        $options = $this->getOptions();
        return $options['name'] ?? null;
    }

    public function getOptions()
    {
        if (is_null($this->_options)) {
            $options = [];
            $rows = $this->getDb()->select("SELECT * FROM ?_option");
            foreach ($rows as $row) {
                $options[$row['option_id']] = $row['data_type'] == 'array' ?
                    unserialize($row['option_value']) : $row['option_value'];
            }
            $this->_options = $options;
        }
        return $this->_options;
    }


}

class Am_Protect_Table_Xenforo extends Am_Protect_Table
{
    /**
     * Escape Identificator
     *
     * @param string $s
     * @return string
     */
    function escapeIdent($s)
    {
        return $this->_db->escape($s, true);
    }

    public function bulkInsert($table, $fields)
    {
        // Insert record only if there is no such record yet.
        $this->_db->query("INSERT IGNORE INTO $table (" . implode(',', array_map([$this, 'escapeIdent'], array_keys($fields))) . ")
            VALUE(" . implode(',', array_map([$this->_db, 'escape'], array_values($fields))) . ")");
    }

    function findOrCreatePermissionCombination($groups, $user_id)
    {
        $needRebuild = false;
        sort($groups, SORT_NUMERIC);
        $user_group_list = implode(',', array_filter($groups));

        $combination_id = $this->findUserPermissionCombination($user_id);

        if ($combination_id) {
            $this->getAdapter()->query("DELETE FROM ?_permission_combination_user_group
                WHERE permission_combination_id=?", $combination_id);

            $this->getAdapter()->query("UPDATE ?_permission_combination SET
                ?a WHERE permission_combination_id=?", [
                    'user_group_list' => $user_group_list,
                    'cache_value' => ''
            ], $combination_id);

            foreach (explode(',', $user_group_list) as $group_id) {
                $this->bulkInsert('?_permission_combination_user_group', [
                    'user_group_id' => $group_id,
                    'permission_combination_id' => $combination_id
                ]);
            }
            $needRebuild = true;
        }

        if (!$combination_id) {
            $combination_id = $this->findPermissionCombination($user_group_list);
        }

        if (!$combination_id) {
            $this->bulkInsert('?_permission_combination', [
                'user_group_list' => $user_group_list,
                'user_id' => 0,
                'cache_value' => ''
            ]);

            $combination_id = $this->findPermissionCombination($user_group_list);
            foreach (explode(',', $user_group_list) as $group_id) {
                $this->bulkInsert('?_permission_combination_user_group', [
                    'user_group_id' => $group_id,
                    'permission_combination_id' => $combination_id
                ]);
            }
            $needRebuild = true;
        }

        //since version 2 we should schedule cache rebuild, before we can just put empty value
        if ($needRebuild && $this->getPlugin()->getConfig('version') == '200') {
            if (!$this->_db->selectCell("SELECT COUNT(*) FROM ?_job WHERE unique_key=?", 'permissionRebuild')) {
                $this->bulkInsert('?_job', [
                    'unique_key' => 'permissionRebuild',
                    'execute_class' => 'XF:PermissionRebuild',
                    'execute_data' => serialize([]),
                    'manual_execute' => 0,
                    'trigger_date' => time()
                ]);
            }
        }

        return $combination_id;
    }

    function findPermissionCombination($user_group_list, $user_id = 0)
    {
        return $this->_db->selectCell("SELECT permission_combination_id
            FROM ?_permission_combination
            WHERE user_id=? AND user_group_list=?", $user_id, $user_group_list);
    }

    function findUserPermissionCombination($user_id)
    {
        return $this->_db->selectCell("SELECT permission_combination_id
            FROM ?_permission_combination
            WHERE user_id=?", $user_id);
    }

    public function updateRelatedTables(Am_Record $record)
    {
        if($this->getPlugin()->getConfig('version', Am_Protect_Xenforo::DEFAULT_VERSION) < 200)
        {
            $profile = [
                'user_id' => $record->pk(),
                'dob_day' => 0,
                'dob_month' => 0,
                'dob_year' => 0,
                'status' => '',
                'status_date' => 0,
                'status_profile_post_id' => 0,
                'signature' => '',
                'homepage' => '',
                'location' => '',
                'occupation' => '',
                'following' => '',
                //'identities' => '',
                'csrf_token' => '',
                'avatar_crop_x' => 0,
                'avatar_crop_y' => 0,
                'about' => ''
            ];

            $option = [
                'user_id' => $record->pk(),
                'show_dob_year' => 1,
                'show_dob_date' => 1,
                'content_show_signature' => 1,
                'receive_admin_email' => 1,
                'email_on_conversation' => 1,
                'is_discouraged' => 0,
                'default_watch_state' => '',
                'alert_optout' => '',
                'enable_rte' => 1
            ];
        }
        else
        {
            $profile = [
                'user_id' => $record->pk(),
                'dob_day' => 0,
                'dob_month' => 0,
                'dob_year' => 0,
                'signature' => '',
                'website' => '',
                'following' => '',
                'ignored' => '',
                'avatar_crop_x' => 0,
                'avatar_crop_y' => 0,
                'about' => '',
                'password_date' => 1
            ];

            $option = [
                'user_id' => $record->pk(),
                'show_dob_year' => 1,
                'show_dob_date' => 1,
                'content_show_signature' => 1,
                'receive_admin_email' => 1,
                'email_on_conversation' => 1,
                'is_discouraged' => 0,
                'creation_watch_state' => '',
                'interaction_watch_state' => '',
                'alert_optout' => '',
                'use_tfa' => 0
            ];
        }

        $privacy = [
            'user_id' => $record->pk(),
            'allow_view_profile' => 'everyone',
            'allow_post_profile' => 'everyone',
            'allow_send_personal_conversation' => 'everyone',
            'allow_view_identities' => 'everyone',
            'allow_receive_news_feed' => 'everyone'
        ];

        $this->bulkInsert('?_user_profile', $profile);
        $this->bulkInsert('?_user_option', $option);
        $this->bulkInsert('?_user_privacy', $privacy);
    }

    function updateFromAmember(Am_Record $record, User $user, $groups)
    {
        // If email was changed through aMember, user state should be changed to valid (if user state was set to email_bounced for example)
        if (strcasecmp($user->email, $record->email) !== 0)
        {
            $record->user_state = 'valid';
        }

        parent::updateFromAmember($record, $user, $groups);
        $this->updateRelatedTables($record);
    }

    public function insertFromAmember(User $user, SavedPass $pass, $groups)
    {
        $record = parent::insertFromAmember($user, $pass, $groups);

        $this->updateRelatedTables($record);
        $this->updatePassword($record, $pass);
    }

    public function setGroups(Am_Record $record, $groups)
    {
        parent::setGroups($record, $groups);
        $record->permission_combination_id = $this->findOrCreatePermissionCombination($groups, $record->pk());
        $record->is_banned = count(array_intersect($groups, $this->_plugin->getBannedGroups())) ? 1 : 0;
        $user = $this->getDi()->userTable->findFirstByLogin($record->username);
        $record->is_admin = ($user && ($is_admin = $user->data()->get('xenforo_admin')) == '1') ? $is_admin :
            (count(array_intersect($groups, $this->_plugin->getAdminGroups())) ? 1 : 0);
        $record->user_group_id = $record->display_style_group_id = array_shift($groups);
        $record->secondary_group_ids = implode(',', $groups);
        $record->updateSelectedFields([
            'is_admin', 'is_banned',
            'user_group_id', 'display_style_group_id', 'secondary_group_ids', 'permission_combination_id'
        ]);

        //set flag is_primary for primary group
        $this->_db->query("UPDATE ?_user_group_relation SET is_primary=1 WHERE user_id=? AND user_group_id=?", $record->user_id, $record->user_group_id);
    }

    public function updatePassword(Am_Record $record, SavedPass $saved)
    {
        $row = $this->_db->selectRow("SELECT * FROM ?_user_authenticate WHERE user_id=?", $record->user_id);
        $version = $this->getPlugin()->getConfig('version', Am_Protect_Xenforo::DEFAULT_VERSION);
        $remember_key = $this->getDi()->security->randomString(40);
        if ($version == 104) {
            $data = [
                'hash' => $saved->pass,
                'salt' => $saved->salt,
                'hashFunc' => 'sha256'
            ];
            $row ?
                    $this->_db->query("UPDATE ?_user_authenticate SET
                    scheme_class=?, data=?, remember_key=?
                    WHERE user_id=?", 'XenForo_Authentication_Core', serialize($data), $remember_key, $record->user_id) :
                    $this->_db->query("INSERT INTO ?_user_authenticate
                    (user_id, scheme_class, data, remember_key)
                    VALUE (?,?,?,?)", $record->user_id, 'XenForo_Authentication_Core', serialize($data), $remember_key)
            ;
        }
        elseif ($version == 130) {
            $data = [
                'hash' => $saved->pass
            ];
            $row ?
                    $this->_db->query("UPDATE ?_user_authenticate SET
                    scheme_class=?, data=?, remember_key=?
                    WHERE user_id=?", 'XenForo_Authentication_Core12', serialize($data), $remember_key, $record->user_id) :
                    $this->_db->query("INSERT INTO ?_user_authenticate
                    (user_id, scheme_class, data, remember_key)
                    VALUE (?,?,?,?)", $record->user_id, 'XenForo_Authentication_Core12', serialize($data), $remember_key)
            ;
        }
        elseif ($version == 200) {
            $data = [
                'hash' => $saved->pass
            ];
            $row ?
                    $this->_db->query("UPDATE ?_user_authenticate SET
                    scheme_class=?, data=?
                    WHERE user_id=?", 'XF:Core12', serialize($data), $record->user_id) :
                    $this->_db->query("INSERT INTO ?_user_authenticate
                    (user_id, scheme_class, data)
                    VALUE (?,?,?)", $record->user_id, 'XF:Core12', serialize($data))
            ;
        }
    }

    function checkPassword(Am_Record $record, User $user, $plaintextPass = null)
    {
        $row = $this->_db->selectRow("SELECT * FROM ?_user_authenticate WHERE user_id=?", $record->user_id);
        if (!$row)
            return false;

        $data = unserialize($row['data']);

        if ($plaintextPass) {
            if ($row['scheme_class'] == 'XenForo_Authentication_vBulletin') {
                return md5(md5($plaintextPass) . $data['salt']) === $data['hash'];
            }
            elseif ($row['scheme_class'] == 'XenForo_Authentication_Core') {
                $salt = $data['salt'];
                return $this->_plugin->cryptPassword($plaintextPass, $salt, $user) === $data['hash'];
            }
            elseif ($row['scheme_class'] == 'XenForo_Authentication_Core12') {
                $phppass = new PasswordHash(10, false);
                return $phppass->CheckPassword($plaintextPass, $data['hash']);
            }
            elseif ($row['scheme_class'] == 'XF:Core12') {
                $phppass = new PasswordHash(10, false);
                return $phppass->CheckPassword($plaintextPass, $data['hash']);
            }
            else {
                return false;
            }
        } else {
            $saved = $this->_plugin->findPassword($user);
            if (!$saved)
                return false;

            $data = unserialize($row['data']);
            return $data['hash'] === $saved->pass;
        }
    }

    public function removeRecord(Am_Record $record)
    {
        $tables = [
            'xf_user', 'xf_user_profile',
            'xf_user_option', 'xf_user_privacy',
            'xf_user_authenticate',
            'xf_user_group_relation',
            'xf_user_trophy', 'xf_user_confirmation',
            'xf_user_external_auth', 'xf_permission_entry',
            'xf_permission_entry_content', 'xf_moderator',
            'xf_moderator_content', 'xf_admin', 'xf_admin_permission_entry',
            'xf_user_ban', 'xf_news_feed', 'xf_user_news_feed_cache',
            'xf_conversation_recipient'
        ];

        if ($this->getPlugin()->getConfig('version', Am_Protect_Xenforo::DEFAULT_VERSION) == 200) {
            $tables[] = 'xf_user_remember';
        }

        foreach ($tables as $table) {
            try {
                $this->_db->query("DELETE FROM ?# WHERE user_id = ?", $table, $record->user_id);
            }
            catch (Am_Exception_PluginDb $e) {
                //for example table does not exist
            }
        }

        $this->_db->query("DELETE FROM ?# WHERE profile_user_id = ?", 'xf_profile_post', $record->user_id);
        $this->_db->query("DELETE FROM ?# WHERE alerted_user_id = ?", 'xf_user_alert', $record->user_id);
        $this->_db->query("DELETE FROM ?# WHERE owner_user_id = ?", 'xf_conversation_user', $record->user_id);
        $this->_db->query("DELETE FROM ?# WHERE content_type = 'user' AND user_id = ?", 'xf_ip', $record->user_id);
    }
}