<?php

use Am\Marketplace\Container;
use Am\Marketplace\Slim\Controller\CategoryTypes;
use Am\Marketplace\Slim\Controller\Config;
use Am\Marketplace\Slim\Controller\DeleteCategory;
use Am\Marketplace\Slim\Controller\DeleteListing;
use Am\Marketplace\Slim\Controller\GetListingInfo;
use Am\Marketplace\Slim\Controller\GetListings;
use Am\Marketplace\Slim\Controller\ListCategories;
use Am\Marketplace\Slim\Controller\Me;
use Am\Marketplace\Slim\Controller\Purchase;
use Am\Marketplace\Slim\Controller\SaveCategory;
use Am\Marketplace\Slim\Controller\SaveListing;
use Am\Marketplace\Slim\Controller\SaveSort;
use Am\Marketplace\Slim\Controller\TransactionChat;
use Am\Marketplace\Slim\Controller\TransactionChatPostMessage;
use Am\Marketplace\Slim\Controller\TransactionComment;
use Am\Marketplace\Slim\Controller\TransactionComplete;
use Am\Marketplace\Slim\Controller\TransactionRevert;
use Am\Marketplace\Slim\Controller\Transactions;
use Am\Marketplace\Slim\Controller\TransactionSent;
use Am\Marketplace\Slim\Controller\Upload as UploadController;
use Am\Marketplace\Slim\Middleware\AdminAuth;
use Am\Marketplace\Slim\Middleware\RequireAdminAuth;
use Am\Marketplace\Slim\Middleware\RequireMarketplaceAccess;
use Am\Marketplace\Slim\Middleware\RequireMarketplaceUserAccess;
use Am\Marketplace\Slim\Middleware\RequireOwnListingID;
use Am\Marketplace\Slim\Middleware\RequireSuperAdmin;
use Am\Marketplace\Slim\Middleware\RequireUserAuth;
use Am\Marketplace\Slim\Response;
use Psr\Http\Message\ServerRequestInterface;
use Slim\App;
use Slim\Factory\AppFactory;
use Slim\Factory\ServerRequestCreatorFactory;
use Slim\Middleware\ContentLengthMiddleware;
use Slim\Middleware\MethodOverrideMiddleware;
use Slim\Psr7\Request;
use Slim\Routing\RouteCollectorProxy;

class Marketplace_AjaxController extends Am_Mvc_Controller
{
    function bootstrap(): App
    {
        AppFactory::setContainer(new Container());
        
        $app = AppFactory::create(new Response\Factory());
        
        $app->setBasePath(REL_ROOT_URL . '/marketplace/ajax');
        $app->add(new ContentLengthMiddleware());
        $app->addBodyParsingMiddleware();
        $app->addRoutingMiddleware();
        $app->add(new MethodOverrideMiddleware());

// Add Error Middleware
        $errorMiddleware = $app->addErrorMiddleware(true, true, true);
        $errorMiddleware->setDefaultErrorHandler(function (
            ServerRequestInterface $request,
            Throwable $exception,
            bool $displayErrorDetails,
            bool $logErrors,
            bool $logErrorDetails
        ) use ($app)
        {
            if ($logErrors) {
                $logger = $app->getContainer()->get('logger');
                $logger->debug($exception->getMessage());
            }
            
            $response = Response::jsonError($exception->getMessage(), 500);
            return $response;
        });
        return $app;
        
    }
    
    function indexAction()
    {
        $app = $this->bootstrap();
        
        $app->group('/admin', function (RouteCollectorProxy $group)
        {
            $group->post('/save-category', SaveCategory::class);
            $group->post('/delete-category', DeleteCategory::class);
            $group->post('/save-sort', SaveSort::class);
        })->add(new RequireSuperAdmin())
            ->add(new RequireAdminAuth());
        
        $app->group('/category', function (RouteCollectorProxy $group)
        {
            $group->post('/list', ListCategories::class);
            $group->post('/types', CategoryTypes::class);
        });
        
        $app->post('/upload', UploadController::class);
        $app->post('/config', Config::class);
        $app->group('/listing', function (RouteCollectorProxy $group)
        {
            
            $group->post('/save', SaveListing::class)
                ->add(new RequireMarketplaceAccess())
                ->add(new RequireUserAuth());
            $group->post('/get', GetListingInfo::class)
                ->add(new RequireOwnListingID())
                ->add(new RequireMarketplaceAccess())
                ->add(new RequireUserAuth());
            $group->post('/delete', DeleteListing::class)
                ->add(new RequireOwnListingID())
                ->add(new RequireMarketplaceAccess())
                ->add(new RequireUserAuth());
            
            $group->post('/purchase', Purchase::class)
                ->add(new RequireMarketplaceUserAccess())
                ->add(new RequireUserAuth());
            
            
            $group->post('/list', GetListings::class);
            
            $group->post('/list/{type:my}', GetListings::class)
                ->add(new RequireMarketplaceAccess())
                ->add(new RequireUserAuth());
            
            
        });
        $app->group('/transaction', function (RouteCollectorProxy $group)
        {
            $group->post('/list', Transactions::class);
            $group->post('/comment', TransactionComment::class);
            $group->post('/complete', TransactionComplete::class);
            $group->post("/sent", TransactionSent::class)->add(new RequireMarketplaceAccess());
            $group->post("/revert", TransactionRevert::class)->add(new RequireMarketplaceAccess());
            $group->group('/chat', function (RouteCollectorProxy $chatGroup)
            {
                $chatGroup->post('/list', TransactionChat::class);
                $chatGroup->post('/post', TransactionChatPostMessage::class);
                
            });
            
        })
            ->add(new RequireMarketplaceUserAccess())
            ->add(new RequireUserAuth());
        
        
        $app->post('/save-listing', SaveListing::class)
            ->add(new RequireMarketplaceAccess())
            ->add(new RequireUserAuth());
        
        
        $app->post('/me', Me::class);
        
        $app->get('/', function (Request $request, Response $response, array $args)
        {
            $response->getBody()->write('Hello World');
            return $response;
        });
        
        $serverRequestCreator = ServerRequestCreatorFactory::create();
        $request = $serverRequestCreator->createServerRequestFromGlobals();
        
        
        $app->run($request->withAttribute('container', $app->getContainer()));
    }
}