import {makeAutoObservable, runInAction} from "mobx";
import {createContext, useContext} from "react";
import {TCategory, TCategoryList, TCategoryListConfig, TCategorySelectOption, TCategoryTreeItem} from "../types";
import {TApiResponse} from "../../types";

export enum TListStatus {
    IDLE,
    LOADING
}

type TSortData = {
    c: string
    s: string
}

export type TCategoryStoreOptions = {
    uri?: string
    type_id?: string
    list?: Array<TCategory>,
    max_upload_width?: number
    max_upload_size?: number

}


export class CategoryStore {
    uri: string = ''
    categories: TCategoryList = {}
    ids: Array<string> = []
    listStatus: TListStatus = TListStatus.IDLE
    currentForm: string | null = null
    categoryTree: Array<TCategoryTreeItem> = []
    categorySelectOptions: Array<TCategorySelectOption> = []
    search: string = "";
    type_id: string = '';
    max_upload_width: number = 200
    max_upload_size: number = 200

    constructor(options: TCategoryStoreOptions) {
        this.uri = options.uri ?? ""
        this.type_id = options.type_id ?? ""
        this.max_upload_size = options.max_upload_size??200
        this.max_upload_width = options.max_upload_width??200
        if (options.list) {
            this.createFromList(options.list);
        }
        makeAutoObservable(this);
    }

    createFromList(list: Array<TCategory>): void {
        this.categories = {};
        this.ids = [];
        list.forEach((category) => {
            this.addCategory(category);
        })
        this._rebuildInternal();

    }


    isFormDisplayed(formId: string) {
        return formId === this.currentForm;
    }

    showForm(formId: string) {
        this.currentForm = formId;
    }

    closeForm() {
        this.currentForm = null;
    }

    getCategoryChildIds(parentId?: string) {
        const search = parentId ?? null
        return this.ids.filter((id) => this.categories[id].parent_id == search);
    }

    buildTree(parentId?: string) {
        const current = parentId ?? null;
        return this.ids.reduce<Array<TCategoryTreeItem>>((prev, cur) => {
            if (this.categories[cur].parent_id === current) {
                prev.push(Object.assign({}, {
                    id: cur,
                    sort: this.categories[cur].sort ?? "0",
                    children: this.buildTree(cur)
                }))
            }
            return prev
        }, []);
    }

    _rebuildInternal() {
        this.categoryTree = this.sortTree(this.buildTree());
        this.categorySelectOptions = this.buildCategorySelect(this.categoryTree);
    }

    _updateSort(data: Array<TCategoryTreeItem>, current: { i: number } = {i: 0}): Array<TSortData> {
        let sort: Array<TSortData> = [];
        data.forEach((el) => {
            el.sort = current.i.toString();
            sort.push({c: el.id, s: el.sort});
            current.i = current.i + 1;
            if (el.children.length > 0) {
                sort = [...sort, ...this._updateSort(el.children, current)];
            }
        });
        return sort;
    }

    async apiMoveUp(list: Array<TCategoryTreeItem>, id: string): Promise<TApiResponse> {
        const ind = list.findIndex((el) => el.id == id);
        if (ind <= 0)
            return Promise.resolve({ok: true} as TApiResponse);
        runInAction(() => {
            let tmp = list[ind];
            list[ind] = list[ind - 1]
            list[ind - 1] = tmp
        })
        return this.apiUpdateSort();
    }

    async apiMoveDown(list: Array<TCategoryTreeItem>, id: string): Promise<TApiResponse> {
        const ind = list.findIndex((el) => el.id == id);
        if (ind === -1 || ind === (list.length - 1))
            return Promise.resolve({ok: true} as TApiResponse);
        runInAction(() => {
            let tmp = list[ind];
            list[ind] = list[ind + 1]
            list[ind + 1] = tmp
        })

        return this.apiUpdateSort();
    }

    async apiUpdateSort(): Promise<TApiResponse> {
        const sort = this._updateSort(this.categoryTree);
        return await this.postData(this.uri + '/admin/save-sort', {sort: sort, type_id: this.type_id}).then()
    }


    sortTree(elements: Array<TCategoryTreeItem>) {
        let els = elements.sort((a, b) => ((a.sort as unknown as number) - (b.sort as unknown as number)))
        els.forEach((el) => el.children.length > 0 && (el.children = this.sortTree(el.children)))
        return els;
    }

    buildCategorySelect(tree: Array<TCategoryTreeItem>, indent: number = 0) {
        return tree.reduce<Array<TCategorySelectOption>>((prev, cur) => {
            prev.push({value: cur.id, label: "".repeat(indent) + this.categories[cur.id].name, indent: indent})
            prev = [...prev, ...this.buildCategorySelect(cur.children, indent + 1)];
            return prev;
        }, []);
    }

    getCategoryOptionById(categoryId?: string): TCategorySelectOption {
        return this.categorySelectOptions.reduce((prev, cur) => cur.value == categoryId ? cur : prev, {} as TCategorySelectOption)
    }

    getCategoryList(config: TCategoryListConfig) {
        const params = Object.assign({}, {parentId: null}, config);
        let ret = {};
        ret = this.ids.reduce<TCategoryList>((prev, cur) => this.categories[cur].parent_id === params.parentId ? Object.assign(prev, {cur: this.categories[cur]}) : prev, {});
    }

    async apiSaveCategory(category: TCategory): Promise<TApiResponse> {
        const response = await this.postData(this.uri + '/admin/save-category', {
            ...category,
            type_id: this.type_id
        }).then((response) => {
            if (response.ok) {
                this.addCategory(response.category as unknown as TCategory);
                this._rebuildInternal();
            }
            return response;
        });
        return response;
    }

    async apiDeleteCategory(category: TCategory): Promise<TApiResponse> {
        const response = await this.postData(this.uri + '/admin/delete-category', {
            id: category.category_id,
            type_id: this.type_id
        }).then((response) => {
            if (response.ok) {
                return this.apiLoad();
            }
            return response;
        });
        return response;
    }

    addCategory(category: TCategory) {
        Object.assign(this.categories, {[category.category_id as string]: category})
        this.ids = Object.keys(this.categories);
    }

    async apiLoad(): Promise<TApiResponse> {
        this.listStatus = TListStatus.LOADING;
        this.categories = {};
        this.ids = [];
        const response = await this.postData(this.uri + '/category/list', {type_id: this.type_id})
            .then((response) => {
                if (response.ok) {
                    runInAction(() => {
                        (response.list as unknown as TCategory[]).forEach((category) => {
                            this.addCategory(category);
                        })
                        this._rebuildInternal();
                    })
                }
                return response;
            }).finally(() => {
                runInAction(() => {
                    this.listStatus = TListStatus.IDLE;

                })
            });

        return response;

    }

    async postData(url = '', data = {}): Promise<TApiResponse> {
        const response = await fetch(url, {
            method: 'POST',
            mode: 'cors',
            cache: 'no-cache',
            credentials: 'same-origin',
            headers: {
                'Content-Type': 'application/json'
            },
            redirect: 'follow',
            referrerPolicy: 'no-referrer',
            body: JSON.stringify(data)
        });

        const json = await response.json();

        if (json.ok === false)
            throw new Error(json.errorMessage);

        return json;
    }

    setSearch(search: string) {
        this.search = search;
    }
    getCategory(id: string)
    {
        return Object.keys(this.categories).reduce<TCategory|null>((carry, item)=>{
            return item==id?this.categories[item] : carry
        }, null);
    }

}


export const CategoryStoreContext = createContext<CategoryStore>({} as any)
export const useCategoryStore = () => useContext(CategoryStoreContext)