import React, {FC} from "react"

import {
    Autocomplete,
    autocompleteClasses,
    Button,
    ClickAwayListener,
    Grid,
    InputBase,
    Menu,
    MenuItem,
    Popper,
    PopperProps,
    styled,
    Typography
} from "@mui/material";
import SearchBar from "./SearchBar";
import KeyboardArrowDownIcon from "@mui/icons-material/KeyboardArrowDown";
import {useListingStore, useMarketplaceStore} from "../Store";
import {observer} from "mobx-react";
import {TCategorySelectOption} from "../../AdminCategory/types";
import {CategoryStore} from "../../AdminCategory/Store";
import {useSearchParams} from "react-router-dom";

type TFilterMenuProps<T> = {
    selected?: T,
    options: Array<{ id: T, title: string }>
    onSelected: (id?: T) => void
    defaultTitle: string
}
const FilterMenu: FC<TFilterMenuProps<number | string>> = (props) => {
    const [anchor, setAnchor] = React.useState<null | HTMLElement>(null);
    const openSort = Boolean(anchor);
    const handleClick = (event: React.MouseEvent<HTMLButtonElement>) => {
        setAnchor(event.currentTarget);
    };
    const handleMenuSelect = (id?: string | number, title?: string) => {
        if (id)
            props.onSelected(id);
        setAnchor(null);
    };
    return <>
        <Button
            id="sort-button"
            aria-controls={openSort ? 'basic-menu' : undefined}
            aria-haspopup="true"
            aria-expanded={openSort ? 'true' : undefined}
            onClick={handleClick}
            size={"small"}
        >
            {props.selected ? props.options.reduce((prev, cur) => ((cur.id == props.selected) ? cur.title : prev), "") : props.defaultTitle}&nbsp;
            <KeyboardArrowDownIcon/>
        </Button>
        <Menu
            id="basic-menu"
            anchorEl={anchor}
            open={openSort}
            onClose={() => handleMenuSelect()}
            MenuListProps={{
                'aria-labelledby': 'sort-button',
            }}
        >
            {props.options.map((el, i) => <MenuItem onClick={() => handleMenuSelect(el.id, el.title)}
                                                    key={i}>{el.title}</MenuItem>)}
        </Menu>

    </>
}
const StyledInput = styled(InputBase)(({theme}) => ({
    padding: 10,
    width: '100%',
    borderBottom: `1px solid ${
        theme.palette.mode === 'light' ? '#eaecef' : '#30363d'
    }`,
    '& input': {
        borderRadius: 4,
        backgroundColor: theme.palette.mode === 'light' ? '#fff' : '#0d1117',
        padding: 8,
        transition: theme.transitions.create(['border-color', 'box-shadow']),
        border: `1px solid ${theme.palette.mode === 'light' ? '#eaecef' : '#30363d'}`,
        fontSize: 14,
        '&:focus': {
            boxShadow: `0px 0px 0px 3px ${
                theme.palette.mode === 'light'
                    ? 'rgba(3, 102, 214, 0.3)'
                    : 'rgb(12, 45, 107)'
            }`,
            borderColor: theme.palette.mode === 'light' ? '#0366d6' : '#388bfd',
        },
    },
}));
const StyledAutocompletePopper = styled('div')(({theme}) => ({
    [`& .${autocompleteClasses.paper}`]: {
        boxShadow: 'none',
        margin: 0,
        color: 'inherit',
        fontSize: 13,
    },
    [`& .${autocompleteClasses.listbox}`]: {
        backgroundColor: theme.palette.mode === 'light' ? '#fff' : '#1c2128',
        padding: 0,
        [`& .${autocompleteClasses.option}`]: {
            minHeight: 'auto',
            alignItems: 'flex-start',
            padding: 8,
            borderBottom: `1px solid  ${
                theme.palette.mode === 'light' ? ' #eaecef' : '#30363d'
            }`,
            '&[aria-selected="true"]': {
                backgroundColor: 'transparent',
            },
            [`&.${autocompleteClasses.focused}, &.${autocompleteClasses.focused}[aria-selected="true"]`]:
                {
                    backgroundColor: theme.palette.action.hover,
                },
        },
    },
    [`&.${autocompleteClasses.popperDisablePortal}`]: {
        position: 'relative',
    },
}));

function PopperComponent(props: PopperProps) {
    const {disablePortal, anchorEl, open, ...other} = props;
    return <StyledAutocompletePopper {...other} />;
}

const StyledPopper = styled(Popper)(({theme}) => ({
    border: `1px solid ${theme.palette.mode === 'light' ? '#e1e4e8' : '#30363d'}`,
    boxShadow: `0 8px 24px ${
        theme.palette.mode === 'light' ? 'rgba(149, 157, 165, 0.2)' : 'rgb(1, 4, 9)'
    }`,
    borderRadius: 6,
    width: 500,
    zIndex: theme.zIndex.modal,
    fontSize: 13,
    color: theme.palette.mode === 'light' ? '#24292e' : '#c9d1d9',
    backgroundColor: theme.palette.mode === 'light' ? '#fff' : '#1c2128',
}));


const CategorySelect: FC<{ title: string, category: CategoryStore, categoryFilter: string | null, onSetFilter: (filter: string) => void }> = observer((props) => {
    const [anchorEl, setAnchorEl] = React.useState(null);

    // @ts-ignore
    const handleCategoryClick = (event) => {
        setAnchorEl(event.currentTarget);
    };

    const handleClose = () => {
        if (anchorEl) {
            // @ts-ignore
            anchorEl.focus();
        }
        setAnchorEl(null);
    };
    const open = Boolean(anchorEl);
    return (
        <>
            <Button
                id="category-button"
                aria-haspopup="true"
                aria-expanded={true}
                disableRipple onClick={handleCategoryClick}
                size={"small"}
            >{props.categoryFilter ? props.category.categorySelectOptions.reduce<string>((p, c) => c.value == props.categoryFilter ? c.label : p, "") : props.title}&nbsp;
                <KeyboardArrowDownIcon/></Button>
            <StyledPopper open={open} anchorEl={anchorEl} placement="bottom-start">
                <ClickAwayListener onClickAway={handleClose}>
                    <Autocomplete
                        open={true}
                        size={"small"}
                        id="listing-category"
                        options={props.category.categorySelectOptions}
                        getOptionLabel={(option) => option.label}
                        PopperComponent={PopperComponent}
                        onChange={(event, value) => props.onSetFilter(value?.value ?? "")}
                        renderOption={(props, option, {selected}) => (
                            <li {...props} >
                                    <span
                                        dangerouslySetInnerHTML={{__html: "&nbsp;&nbsp;&nbsp;".repeat(option.indent ?? 0)}}/>
                                {option.label}
                            </li>
                        )}
                        style={{width: 500}}
                        renderInput={(params) => (
                            <><StyledInput
                                ref={params.InputProps.ref}
                                inputProps={params.inputProps}
                                autoFocus
                                placeholder={"Select " + props.title}
                            /><Button onClick={() => {
                                setAnchorEl(null);
                                props.onSetFilter("")
                            }}>Show all</Button>
                            </>
                        )}

                        value={props.category.categorySelectOptions.reduce<TCategorySelectOption | null>((p, c) => c.value == props.categoryFilter ? c : p, null)}
                    /></ClickAwayListener></StyledPopper>

        </>
    )
})
const FilterBar = observer(() => {
    const listingStore = useListingStore();
    const store = useMarketplaceStore();
    const [searchParams, setUrlSearchParams] = useSearchParams();
    const setFilter = (filter: Record<string, string>) => {
        const oldParams = {};
        searchParams.forEach((value, key) => {
            Object.assign(oldParams, {[key]: value});
        })

        let newParams = Object.assign({}, oldParams, filter);
        setUrlSearchParams(newParams);
    }

    return <>
        <Grid item container xs={6} flexDirection={"row"} justifyContent={"flex-end"}>
            {store.categories.types.map((type, i) => <Grid item key={i}>
                    <CategorySelect
                        category={store.categories.getCategoriesByType(type.type_id)}
                        categoryFilter={searchParams.get('cat-' + type.slug)}
                        onSetFilter={(category) => setFilter({['cat-' + type.slug]: category})}
                        title={type.title}/>
                </Grid>
            )}
        </Grid>
        <Grid container item xs={12} sx={{paddingTop: '1em'}} justifyContent={"space-between"}
              alignItems={"center"} direction={"row"}>
            <Grid item container xs={6} alignItems={"center"}>
                <Typography variant={"body2"}>Found:&nbsp;{listingStore.totalRecords}&nbsp;items</Typography>
                <FilterMenu
                    options={listingStore.sortOptions}
                    selected={listingStore.sort}
                    defaultTitle={"Sort By"}
                    onSelected={(id) => listingStore.setSort(id as string)}
                />

            </Grid>

            <Grid item container xs={6} justifyContent={"flex-end"}>
                <Grid item>
                    <SearchBar onSearchChange={(text) => setFilter({text: text})}
                               search={listingStore.getFilter(searchParams).text}
                               filterIcon={false}
                    />

                </Grid>
            </Grid>

        </Grid>
    </>
})
export default FilterBar;

