import React, {FC, useEffect, useState} from "react";
import {
    Alert,
    Autocomplete,
    Box,
    Button,
    CircularProgress,
    FormControl,
    Grid,
    InputLabel,
    MenuItem,
    Select,
    TextField, Typography
} from "@mui/material";
import {TListingFormProps, TListingFormValue} from "../../types";
import noop from "../../../Utils/noop";
import {TCategorySelectOption} from "../../../AdminCategory/types";
import ImageUpload from "../ImageUpload";
import {useMarketplaceStore} from "../../Store";
import Img from "./Img";
import ListingImage from "./Image";


const Form: FC<TListingFormProps> = (props) => {
    const [formValue, setFormValue] = useState<TListingFormValue>(props.listing ?? {
        status: 'draft'
    } as TListingFormValue);
    const onCancel = props.onCancel ?? noop
    const onSubmit = props.onSubmit ?? noop
    const store = useMarketplaceStore();
    const setCategory = (option: TCategorySelectOption | null, type_id: string) => {
        let category_id = formValue.category_id ?? {};
        category_id = Object.assign(category_id, {[type_id]: option?.value});
        setFormValue({...formValue, category_id: category_id});
    }
    const isDisabled = props.isDisabled ?? false
    const isError = (element: keyof TListingFormValue | undefined = undefined) => (props.error !== undefined && element !== undefined && typeof props.error === 'object' && props.error.hasOwnProperty(element)) || (typeof props.error === 'string' && element === undefined)
    const getError = (element: keyof TListingFormValue | undefined = undefined) => {
        if (element === undefined && typeof props.error === 'string')
            return props.error

        if (element !== undefined && typeof props.error === 'object' && props.error.hasOwnProperty(element))
            return props.error[element as keyof typeof props.error];

        return undefined
    }

    const [categoryImages, setCategoryImages] = useState<Array<{image: string, imageUrl:string}>>([]);
    useEffect(()=>{
        let images:Array<{image: string, imageUrl:string}> = [];
        if(formValue.category_id)
        {
            Object.keys(formValue.category_id).forEach((el) => {
                if(formValue.category_id?.hasOwnProperty(el)){
                    let cat = store.categories.getCategoriesByType(el).getCategory(formValue.category_id[el]);
                    if(cat && cat.imageUrl&&cat.image){
                        images = [...images, {image: cat.image, imageUrl: cat.imageUrl}];
                    }
                }
            })
        }
        setCategoryImages(images);
    }, [formValue]);

    return (
        <form id={"add-form"}>
            <Grid container direction={'column'} spacing={2}>
                {isError() && <Grid item>
                    <Alert severity={'error'}>
                        {getError()}
                    </Alert>
                </Grid>
                }
                <Grid item>
                    <TextField fullWidth id="listing-title" aria-describedby="listing-title-text"
                               variant={"outlined"}
                               label={"Listing Title"} required
                               helperText={isError('title') ? getError('title') : "max 32 characters"}
                               onChange={event => {
                                   setFormValue({...formValue, title: event.target.value})
                               }}
                               disabled={props.isLoading || props.isDisabled}
                               error={isError('title')}
                               defaultValue={formValue.title}
                    />
                </Grid>
                <Grid item>
                    <TextField id={"listing-description"} fullWidth multiline rows={5} label="Description"
                               variant="outlined"
                               size={"small"}
                               helperText={isError('description') ? getError('description') : "HTML is not allowed"}
                               onChange={(event) => setFormValue({...formValue, description: event.target.value})}
                               disabled={props.isLoading || props.isDisabled}
                               error={isError('description')}
                               defaultValue={formValue.description}
                    />
                </Grid>
                <Grid item>
                    <Typography variant={"body1"}>Listing Categories</Typography>
                </Grid>
                {store.categories.types.map((type, i) => <Grid item key={i}>
                        <Autocomplete
                            size={"small"}
                            id={"listing-category-" + type.type_id}
                            options={store.categories.getCategoriesByType(type.type_id).categorySelectOptions}
                            getOptionLabel={(option) => option.label}
                            onChange={(event, value) => setCategory(value, type.type_id)}
                            renderOption={(props, option, {selected}) => (
                                <li {...props} >
                                    <span
                                        dangerouslySetInnerHTML={{__html: "&nbsp;&nbsp;&nbsp;".repeat(option.indent ?? 0)}}/>
                                    {option.label}
                                </li>
                            )}
                            style={{width: 500}}
                            renderInput={(params) => (
                                <TextField {...params} label={type.title} placeholder="Please select"/>
                            )}
                            disabled={props.isLoading || props.isDisabled}
                            value={store.categories.getCategoriesByType(type.type_id).categorySelectOptions.reduce<TCategorySelectOption | null>((p, c) => Object.values(formValue.category_id ?? {}).includes(c.value) ? c : p, null)}
                        />
                    </Grid>
                )}

                <Grid item>
                    <Typography variant={"body1"}>Listing Image</Typography>
                </Grid>
                {categoryImages.length > 0 && !formValue.image && <Grid item container direction={"row"}>
                    {categoryImages.map((item, key)=><Grid item key={key}>
                        <Box onClick={()=>setFormValue({...formValue, image: item.image, imageUrl: item.imageUrl})} sx={{cursor: "pointer"}}><ListingImage src={item.imageUrl} maxHeight={"100px"}
                                                                                        maxWidth={"100px"}/>
                        </Box>
                    </Grid>)}
                </Grid>}

                <Grid item>
                    {props.isDisabled ?
                        <Img src={formValue?.imageUrl} width={store.config.maxUploadWidth}/> : <ImageUpload
                            url={formValue?.imageUrl}
                            onSetFile={(file) => setFormValue({...formValue, upload: file})}
                            onDeleteFile={() => setFormValue({...formValue, image: undefined, imageUrl: undefined})}
                            {...store.config}/>}

                </Grid>

                <Grid item>
                    <FormControl>
                        <InputLabel id="listing-type-label">Item Type</InputLabel>
                        <Select
                            labelId="listing-type-label"
                            id="listing-type"
                            defaultValue={formValue.listing_type ?? 'regular'}
                            label="Item Type"
                            size={"small"}
                            disabled={props.isLoading || props.isDisabled}
                            onChange={(event) => setFormValue({...formValue, listing_type: event.target.value})}
                        >
                            <MenuItem value={"regular"}>Regular</MenuItem>
                            <MenuItem value={"serials"}>Serial Code</MenuItem>
                        </Select>
                    </FormControl>
                </Grid>

                {props.user.canSetQty && formValue.listing_type !== 'serials' &&
                    <Grid item>
                        <TextField id={"listing-qty"} type="number" size={"small"} label={"Quantity"}
                                   sx={{width: '12ch'}}
                                   disabled={props.isLoading || props.isDisabled}
                                   error={isError('qty')}
                                   helperText={getError('qty')}
                                   onChange={(event) => setFormValue({...formValue, qty: parseInt(event.target.value)})}
                                   defaultValue={formValue.qty}

                        />
                    </Grid>}
                {formValue.listing_type == 'serials' && <Grid item>
                    <TextField id={"listing-serials"} fullWidth multiline rows={5} label="Serial Codes"
                               variant="outlined"
                               size={"small"}
                               helperText={isError('serial_codes') ? getError('serial_codes') :
                                   <span>Include Serial Codes in text box above(one per line). Serial code will be displayed at the bottom of "Download Instructons"</span>}
                               disabled={props.isLoading || props.isDisabled}
                               onChange={(event) => setFormValue({...formValue, serial_codes: event.target.value})}
                               error={isError('serial_codes')}
                               value={formValue.serial_codes}
                    />

                </Grid>}

                <Grid item>
                    <TextField id={"listing-instructions"} fullWidth multiline rows={5} label="Download Instructions"
                               variant="outlined"
                               size={"small"}
                               helperText={isError('instructions') ? getError('instructions') :
                                   <span>Instructions will be displayed to buyer after Safe Deal is initiated</span>}
                               disabled={props.isLoading || props.isDisabled}
                               onChange={(event) => setFormValue({...formValue, instructions: event.target.value})}
                               error={isError('instructions')}
                               value={formValue.instructions}
                    />

                </Grid>

                <Grid item>
                    <TextField id={"listing-price"} type="number" size={"medium"} label={"Item Price"}
                               sx={{width: '25ch'}}
                               helperText={isError('price') ? getError('price') : "Price in credits"} required
                               disabled={props.isLoading || props.isDisabled}
                               onChange={(event) => setFormValue({...formValue, price: parseInt(event.target.value)})}
                               error={isError('price')}
                               value={formValue.price}

                    />
                </Grid>
                {formValue.listing_type == 'regular' && <Grid item>
                    <TextField id={"listing-delivery"} type="number" size={"medium"} label={"Delivery Timeframe"}
                               sx={{width: '25ch'}}
                               helperText={isError('delivery') ? getError('delivery') : "Delivery Timeframe in Hours"}
                               required
                               disabled={props.isLoading || props.isDisabled}
                               onChange={(event) => setFormValue({
                                   ...formValue,
                                   delivery: parseInt(event.target.value)
                               })}
                               error={isError('delivery')}
                               value={formValue.delivery}

                    />
                </Grid>}

                <Grid item>
                    <FormControl>
                        <InputLabel id="listing-status-label">Status</InputLabel>
                        <Select
                            labelId="listing-status-label"
                            id="listing-status"
                            defaultValue={formValue.status ?? 'draft'}
                            label="Status"
                            size={"small"}
                            disabled={props.isLoading || props.isDisabled}
                            onChange={(event) => setFormValue({...formValue, status: event.target.value})}
                        >
                            <MenuItem value={"draft"}>Draft</MenuItem>
                            <MenuItem value={'published'}>Published</MenuItem>
                        </Select>
                    </FormControl>
                </Grid>


                <Grid item>
                    <Grid container spacing={1}>
                        <Grid item>
                            <Box sx={{position: 'relative'}}>
                                <Button
                                    variant="contained"
                                    disabled={props.isLoading || props.isDisabled}
                                    onClick={() => onSubmit(formValue)}
                                >
                                    SAVE
                                </Button>
                                {props.isLoading && (
                                    <CircularProgress
                                        size={24}
                                        sx={{
                                            color: "primary",
                                            position: 'absolute',
                                            top: '50%',
                                            left: '50%',
                                            marginTop: '-12px',
                                            marginLeft: '-12px',
                                        }}
                                    />
                                )}
                            </Box>
                        </Grid>
                        <Grid item>
                            <Button variant={"outlined"} disabled={props.isLoading}
                                    onClick={event => onCancel()}>CANCEL</Button>
                        </Grid>

                    </Grid>

                </Grid>


            </Grid>
        </form>);
}

export default Form;