<?php

/**
 * @title Helpdesk Tickets By Email
 * @desc Automatically convert all emails from your customers to tickets in helpdesk
 * @am_plugin_api 6.0
 */
class Am_Plugin_HelpdeskTicketsByEmail extends Am_Plugin
{
    const PLUGIN_STATUS = self::STATUS_PRODUCTION;
    const PLUGIN_COMM = self::COMM_COMMERCIAL;
    const PLUGIN_REVISION = '6.3.21';

    protected $_configPrefix = 'misc.';

    static function getDbXml()
    {
        return <<<CUT
<schema version="4.0.0">
    <table name="helpdesk_mail_account">
        <field name="mail_account_id" type="int" notnull="1" extra="auto_increment"/>
        <field name="type" type="enum" len="'pop3','pop3-ssl','imap','imap-ssl'"/>
        <field name="user" type="varchar" len="255"/>
        <field name="pwd" type="varchar" len="255"/>
        <field name="host" type="varchar" len="255"/>
        <field name="port" type="int" unsigned="1" notnull="0"/>
        <index name="PRIMARY" unique="1">
            <field name="mail_account_id" />
        </index>
    </table>
    <table name="helpdesk_message">
        <field name="email_message_id" type="varchar" len="255" notnull="0"/>
    </table>
</schema>
CUT;
    }

    static function getEtXml()
    {
        return <<<CUT
<table_data name="email_template">
    <row type="email_template">
        <field name="name">misc.helpdesk-tickets-by-email.new_message</field>
        <field name="email_template_layout_id">1</field>
        <field name="lang">en</field>
        <field name="format">text</field>
        <field name="subject">%site_title%: New Message for Your Ticket [#%ticket.ticket_mask%] %ticket.subject%</field>
        <field name="txt">

%message.content%

        </field>
    </row>
</table_data>
CUT;
    }

    function init()
    {
        if ($this->isConfigured()) {
            $this->getDi()->config->set('helpdesk.notify_new_message', 0);
        }
    }

    public function _initSetupForm(Am_Form_Setup $form)
    {
        $form->setTitle('Helpdesk Tickets By Email');

        if (!extension_loaded("imap")) {
            $form->addProlog(<<<'CUT'
<div class="admin-warnings">
    <ul>
        <li>This plugin require <strong>imap</strong> PHP extensions.</li>
    </ul>
</div>
CUT
);
        }

        $l_m = Am_Html::escape(___('manage'));

        $form->addHtml()
            ->setLabel(___('Mail Accounts'))
            ->setHtml("<a href='javascript:;' class='link local' onclick='amOpenMailAccounts();' title='Manage Mail Accounts'>{$l_m}</a>");

        $cron_url = $this->getDi()->surl("misc/{$this->getId()}/cron");

        $form->addHtml(null, ['class' => 'am-row-wide'])
            ->setLabel('Cron Configuration')
            ->setHtml(<<<CUT
<p>It is required to setup a cron job to run each 10 minutes to check email accounts</p>
<div style="background: #eee; padding: .2em; margin: 1em 0"><pre>*/10 * * * * /usr/bin/curl {$cron_url}</pre></div>
<p>It is not recommended to check it more frequently because mail service provider (eg.: gmail) can prohibit it.</p>
CUT
);

        $form->addElement('email_link', 'new_message')
            ->setLabel(___("Email Message to User\n" .
                "aMember will email a notification to user " .
                "each time admin responds to a user ticket, user " .
                "can reply to this message from his mail agent " .
                "directly, response will be added to ticket automatically"));

        $manage_account_url = json_encode($this->getDi()->url('admin-helpdesk-mail-account', false));

        $form->addEpilog(<<<CUT
<div id="mail-accounts" style="display: none"></div>
<script type="text/javascript">
function amOpenMailAccounts()
{
    jQuery("#mail-accounts").load({$manage_account_url}
        ,function(){
            jQuery("#mail-accounts").dialog({
                autoOpen: true
                ,width: 800
                ,height: 600
                ,closeOnEscape: true
                ,title: "Manage Mail Accounts"
                ,modal: true
            });
            jQuery("#mail-accounts .am-grid-wrap").ngrid();
        }
    );
}
</script>
CUT
);
    }

    function isConfigured()
    {
        return extension_loaded("imap")
            && $this->getDi()->modules->isEnabled('helpdesk');
    }

    function onSetupEmailTemplateTypes(Am_Event $e)
    {
        $ticket = [
            'ticket.ticket_mask' => 'Ticket Mask',
            'ticket.subject' => 'Ticket Subject',
        ];

        $m = $this->getDi()->modules->loadGet('helpdesk');
        $from = $m->getConfig('email_from') ?
            [$m->getConfig('email_from'), $m->getConfig('email_name')] :
            null;

        $e->addReturn([
            'id' => 'misc.helpdesk-tickets-by-email.new_message',
            'from' => $from,
            'title' => 'Notify New Message',
            'mailPeriodic' => Am_Mail::USER_REQUESTED,
            'vars' => $ticket + ['url' => 'Url of Page with Message', 'user'],
        ], 'misc.helpdesk-tickets-by-email.new_message');
    }

    function directAction($request, $response, $invokeArgs)
    {
        switch ($request->getActionName()) {
            case 'cron' :
                $this->checkMailAccounts();
                break;
            default:
                break;
        }
    }

    function onHelpdeskMessageBeforeInsert(Am_Event $e)
    {
        $message = $e->getMessage();
        if ($message->type == 'comment') return;
        if (empty($message->admin_id)) return;

        $ticket = $message->getTicket();
        $user = $ticket->getUser();

        $et = Am_Mail_Template::load('misc.helpdesk-tickets-by-email.new_message', $user->lang);
        $et->setMessage($message);
        $et->setTicket($ticket);
        $et->setUser($user);
        $et->setUrl($this->getDi()->surl('helpdesk/ticket/'.$message->getTicket()->ticket_mask, false));
        $message->email_message_id = $et->getMail()->setMessageId()->getMessageId();

        //find last users message in thread
        list($lastMessage) = $this->getDi()->helpdeskMessageTable->selectObjects(<<<CUT
                SELECT * FROM ?_helpdesk_message
                    WHERE ticket_id=?
                        AND admin_id IS NULL
                        AND email_message_id IS NOT NULL
                    ORDER BY dattm DESC
                    LIMIT 1;
CUT
        , $ticket->pk());

        if ($lastMessage && $lastMessage->email_message_id) {
            $et->getMail()->addHeader('in-reply-to', "<{$lastMessage->email_message_id}>");
            $et->getMail()->addHeader('references', "<{$lastMessage->email_message_id}>");
        }

        if ($_ = $message->loadGetAttachments()) {
            foreach ($_ as $file) {
                $et->getMail()->createAttachment(
                    file_get_contents($file->getFullPath()),
                    $file->getType(),
                    Zend_Mime::DISPOSITION_ATTACHMENT,
                    Zend_Mime::ENCODING_BASE64,
                    $file->getName()
                );
            }
        }

        $et->send($user);
    }

    function onHourly(Am_Event $e)
    {
        $this->checkMailAccounts();
    }

    function checkMailAccounts()
    {
        if (!$this->getDi()->db->selectCell("SELECT GET_LOCK(?, 0)", $this->getLockId())) {
            return;
        }

//        $this->parseMessage(file_get_contents($this->getDi()->data_dir . '/last-helpdesk-message.msg'));
//        return;

        foreach ($this->getDi()->helpdeskMailAccountTable->findBy() as $ma) {
            if (!$mb = $this->connect($ma, $error)) {
                $this->getDi()->logger->warning(
                    'Can not connect to mail account {user} ({type}): {error}',
                    ['user' => $ma->user, 'type' => $ma->type, 'error' => $error]
                );
                continue; //next account
            }

            $_ = imap_check($mb);
            $err = imap_errors(); //avoid notice in case of mailbox is empty
            for ($i=1; $i<=$_->Nmsgs; $i++) {
                $headers = imap_fetchheader($mb, $i);
                $body = imap_body($mb, $i);

                //save to FS
                file_put_contents($this->getDi()->data_dir . '/last-helpdesk-message.msg', "{$headers}\r\n\r\n{$body}");

                $this->parseMessage($headers . Zend_Mime::LINEEND . $body);

                imap_delete($mb, $i);
                imap_expunge($mb);
            }
        }
    }

    protected function getUser($from)
    {
        if ($u = $this->getDi()->userTable->findFirstByEmail($from)) {
            return $u;
        }

        $user = $this->getDi()->userRecord;
        $user->email = $from;
        $user->generateLogin();
        $user->generatePassword();
        $user->save();

        return $user;
    }

    protected function getTicket($in_reply_to, $user, $subject, & $is_ticket_created)
    {
        if ($in_reply_to &&
            ($msg = $this->getDi()->helpdeskMessageTable->findFirstBy(['email_message_id' => $in_reply_to]))) {

            $is_ticket_created = false;
            return $msg->getTicket();
        }

        $ticket = $this->getDi()->helpdeskTicketRecord;
        $ticket->user_id = $user->pk();
        $ticket->subject = $subject;
        $ticket->status = HelpdeskTicket::STATUS_NEW;
        $ticket->save();

        $is_ticket_created = true;

        return $ticket;
    }

    function parseMessage($msg)
    {
        $part = new Zend_Mail_Part(['raw' => $msg]);

        $from = $this->filterEmail($this->getHeader($part, 'from'));
        $in_reply_to = $this->filterInReplyTo($this->getHeader($part, 'in-reply-to'));
        $subject = $this->getHeader($part, 'subject') ?: '(No Subject)';
        $message_id = $this->filterMessageId($this->getHeader($part, 'message-id'));

        $user = $this->getUser($from);
        $ticket = $this->getTicket($in_reply_to, $user, $subject, $is_ticket_created);

        //message
        $message = $this->getDi()->helpdeskMessageRecord;
        $message->type = 'message';
        $message->ticket_id = $ticket->pk();
        $message->dattm = $this->getDi()->sqlDateTime;
        $message->email_message_id = $message_id;

        $attachments = [];

        $canHasAttachments = false;
        list($type) = explode(";", $this->getHeader($part, 'content-type'));
        if ($type == 'multipart/alternative') {
            $msgPart = $part->getPart(1); //text part
            $attachments[] = $this->createAttachment($user, 'message.html', $this->getContent($part->getPart(2))); //html part
        } else {
            $msgPart = $part->isMultipart() ? $part->getPart(1) : $part;
            if ($msgPart->isMultipart()) {
                $_ = $msgPart;
                $msgPart = $_->getPart(1); //text part
                $attachments[] = $this->createAttachment($user, 'message.html', $this->getContent($_->getPart(2))); //html part
            }
            $canHasAttachments = true;
        }

        list($type) = explode(";", $this->getHeader($msgPart, 'content-type'));

        $content = $this->getContent($msgPart);

        switch ($type) {
            case 'text/plain':
                $message->content = $content;
                break;
            case 'text/html':
                $message->content = strip_tags($content);
                $attachments[] = $this->createAttachment($user, 'message.html', $content);
                break;
        }

        //attachments
        if ($canHasAttachments) {
            if ($part->isMultipart()) {
                for ($i = 2; $i <= $part->countParts(); $i++) {
                    $attPart = $part->getPart($i);

                    $disposition = $this->getHeader($attPart, 'content-disposition');
                    preg_match('/filename="(.*)"/', $disposition, $m);
                    $filename = $m[1];

                    $content = $this->getContent($attPart);
                    $attachments[] = $this->createAttachment($user, $filename, $content);
                }
            }
        }

        if ($attachments) {
            $message->setAttachments($attachments);
        }

        $message->save();

        $strategy = new Am_Helpdesk_Strategy_User($this->getDi(), $user->pk());

        if ($is_ticket_created) {
            $strategy->onAfterInsertTicket($ticket, $message);
        } else {
            $ticket->status = $strategy->getTicketStatusAfterReply($message);
        }
        $strategy->onAfterInsertMessage($message, $ticket);

        $ticket->updated = $this->getDi()->sqlDateTime;
        $ticket->save();
    }

    function filterEmail($header_content_line)
    {
        if (preg_match('/(.*)<(.*)>/', $header_content_line, $m)) {
            $email = trim($m[2]);
        } else {
            $email = trim($header_content_line);
        }

        return $email;
    }

    function filterInReplyTo($header_content_line)
    {
        if (preg_match('/(.*)<(.*)>/', $header_content_line, $m)) {
            $value = trim($m[2]);
        } else {
            $value = trim($header_content_line);
        }

        return $value;
    }

    function filterMessageId($header_content_line)
    {
        return trim(str_replace(['<', '>'], '', $header_content_line));
    }

    function getContent($part)
    {
        $content = $part->getContent();
        switch($this->getHeader($part, 'content-transfer-encoding')) {
            case 'quoted-printable':
                return quoted_printable_decode($content);
            case 'base64':
                return base64_decode($content);
            default:
                return $content;
        }
    }

    function createAttachment($user, $filename, $content)
    {
        $upload = $this->getDi()->uploadRecord;
        $upload->prefix = Bootstrap_Helpdesk::ATTACHMENT_UPLOAD_PREFIX;
        $upload->user_id = $user->pk();
        $upload->setName($filename);
        $upload->mime = Upload::getMimeType($filename);

        do {
            $upload->path = '.' . $upload->prefix . '.' . uniqid();
            $fullPath = $upload->getFullPath();
        } while (file_exists($fullPath));

        file_put_contents($fullPath, $content);
        $upload->save();

       return $upload->pk();
    }

    function getHeader($part, $header)
    {
        $_ = $part->headerExists($header) ? $part->getHeader($header, 'string') : null;
        if (strpos($_, '=?') === 0) {
            $_ = mb_decode_mimeheader($_);
        }
        return $_;
    }

    function connect($ma, & $error = null)
    {
        switch($ma->type) {
            case 'pop3':
                $_ = sprintf("{%s:%d/pop3/notls}INBOX", $ma->host, $ma->port ?: 110);
                break;
            case 'pop3-ssl':
                $_ = sprintf("{%s:%d/pop3/ssl/novalidate-cert}INBOX", $ma->host, $ma->port ?: 995);
                break;
            case 'imap':
                $_ = sprintf("{%s:%d/notls}INBOX", $ma->host, $ma->port ?: 143);
                break;
            case 'imap-ssl':
                $_ = sprintf("{%s:%d/imap/ssl/novalidate-cert}INBOX", $ma->host, $ma->port ?: 993);
                break;
        }

        if (!$imap = @imap_open($_, $ma->user, $ma->pwd)) {
            $error = imap_last_error();
        }

        return $imap;
    }

    public function getLockId()
    {
        return 'cma-lock-' . md5(ROOT_SURL);
    }

    public function getReadme()
    {
        $url = $this->getDi()->url('admin-setup/helpdesk');
        $email = $this->getDi()->modules->isEnabled('helpdesk') ?
            $this->getDi()->modules->loadGet('helpdesk')->getConfig('email_from',
                $this->getDi()->config->get('admin_email_from',
                $this->getDi()->config->get('admin_email'))) :
            $this->getDi()->config->get('admin_email_from',
                $this->getDi()->config->get('admin_email'));

        return <<<CUT
Please check <strong>Outgoing Email Address</strong> ({$email})
within <a href="{$url}" class="link">helpdesk module</a> configuration. This address must
delivers emails to one of configured Mail Accounts above.
Otherwise you will not receive replies.
CUT;
    }
}

class AdminHelpdeskMailAccountController extends Am_Mvc_Controller_Grid
{
    function checkAdminPermissions(Admin $admin)
    {
        return $admin->hasPermission(Am_Auth_Admin::PERM_SETUP);
    }

    function createGrid()
    {
        $q = new Am_Query($this->getDi()->helpdeskMailAccountTable);
        $grid = new Am_Grid_Editable('_hma', ___('Mail Accounts'), $q, $this->getRequest(), $this->getView());
        $grid->actionGet('insert')->setTarget(null);
        $grid->addField('user', ___('Account'))
            ->setRenderFunction(function ($r, $fn, $g, $fo) {
                return $g->renderTd(sprintf("%s (%s)", $r->user, $r->type));
            });

        $grid->setForm([$this, 'createForm']);
        $grid->setRecordTitle(___('Mail Account'));

        $grid->addCallback(Am_Grid_Editable::CB_VALUES_FROM_FORM, function(& $v, $r) {
            $v['port'] = $v['port'] ?: null;
        });

        $grid->isAjax(false); //to open it in popup
        return $grid;
    }

    function createForm($g)
    {
        $form = new Am_Form_Admin('helpdesk-mail-account');

        $form->addSelect('type')
            ->setLabel(___('Protocol'))
            ->loadOptions([
                'pop3' => 'POP3',
                'pop3-ssl' => 'POP3-SSL',
                'imap' => 'IMAP',
                'imap-ssl' => 'IMAP-SSL',
            ]);

        $form->addText('host', ['class' => 'am-el-wide'])
            ->setLabel(___('Host'));

        $form->addText('port', ['placeholder' => ___('leave blank for default'), 'size' => 5])
            ->setLabel(___('Port'));

        $form->addText('user', ['class' => 'am-el-wide'])
            ->setLabel(___('Username'));

        $form->addPassword('pwd')
            ->setLabel(___('Password'));

        $form->addScript()
            ->setScript(<<<CUT
jQuery(function() {
    const ports = {
        'pop3': 110,
        'pop3-ssl': 995,
        'imap': 143,
        'imap-ssl': 993,
    }
    jQuery('[name=type]').change(function(){
        jQuery('[name=port]').attr('placeholder', ports[jQuery(this).val()]);
    }).change();
});

CUT
);

        $form->addRule('callback2', null, [$this, 'checkMA']);

        return $form;
    }

    function checkMA($v)
    {
        $ma = $this->getDi()->helpdeskMailAccountRecord;
        $ma->setForInsert($v);

        if (!$this->getDi()->plugins_misc->loadGet('helpdesk-tickets-by-email')->connect($ma, $error, $cn)) {
            return ___('Can not connect to mail account: %s', $error);
        }
    }
}

class HelpdeskMailAccount extends Am_Record {}

class HelpdeskMailAccountTable extends Am_Table
{
    protected $_table = '?_helpdesk_mail_account';
    protected $_key = 'mail_account_id';
    protected $_recordClass = 'HelpdeskMailAccount';
}