<?php

/**
 * @am_plugin_api 6.0
*/
class Am_Plugin_Intercom extends Am_Plugin
{
    const PLUGIN_STATUS = self::STATUS_BETA;
    const PLUGIN_COMM = self::COMM_COMMERCIAL;
    const PLUGIN_REVISION = '6.3.21';
    const API_ENDPOINT = 'https://api.intercom.io/';
    const DATA_KEY = 'intercom-id';

    protected $_configPrefix = 'misc.';
    protected $_isDebug = false;

    function _initSetupForm(Am_Form_Setup $form)
    {
        $form->addText('app_id')->setLabel('App Id');
        $form->addPassword('pat', ['class' => 'am-el-wide'])->setLabel('Personal Access Token (PAT)');
        $form->addAdvCheckbox('secure', ['id' => 'intercom-secure'])
            ->setLabel("Enable secure mode\n" .
                'Secure mode allows you to make sure that conversations between you and your users are kept private. <a href="https://docs.intercom.io/configuring-for-your-product-or-site/enable-secure-mode" class="link" target="_blank" rel="noreferrer">Learn more about Secure Mode</a>');
        $form->addPassword('secret', ['class' => 'am-el-wide'])
            ->setLabel('Secret Key');
        $form->addAdvCheckbox('w_not_logged_in')
            ->setLabel('Add widget for not logged in users');
        $form->addAdvCheckbox('w_logged_in')
            ->setLabel('Add widget for logged in users');
        $form->addScript()
            ->setScript(<<<CUT
jQuery(function($){
    jQuery('#intercom-secure').change(function(){
        $('[name$=secret]').closest('.am-row').toggle(this.checked);
    }).change();
})
CUT
                );
        return $form;
    }

    function init()
    {
        if ($this->isConfigured() && $w = $this->getWidget()) {
            $this->getDi()->view->placeholder('body-finish')->append($w);
        }
    }

    function isConfigured()
    {
        return (bool)$this->getConfig('app_id');
    }

    function onAdminWarnings(Am_Event $e)
    {
        //api_key was depricated
        //https://medium.com/intercom-developers/upcoming-deprecation-of-api-keys-1d2f9a01b91e
        //https://developers.intercom.com/v2.0/docs/personal-access-tokens
        if ($this->getConfig('api_key') && !$this->getConfig('pat')) {
            $e->addReturn('Intercom: API Keys is depricated. Please fill in PAT (Personal Access Token) within plugin configuration');
        }
    }

    public function onInitFinished(Am_Event $e)
    {
        if ($this->getDi('pat')) {
            $this->getDi()->hook->add(array(
                Am_Event::USER_AFTER_INSERT,
                Am_Event::USER_AFTER_UPDATE,
                Am_Event::SUBSCRIPTION_CHANGED), array($this, 'syncUser'));
            $this->getDi()->hook->add(Am_Event::USER_BEFORE_DELETE, array($this, 'delUser'));
        }
    }

    public function onRebuild(Am_Event_Rebuild $event)
    {
        $batch = new Am_BatchProcessor(array($this, 'batchProcess'));
        $context = $event->getDoneString();
        if ($batch->run($context)) {
            $event->setDone();
        } else {
            $event->setDoneString($context);
        }
    }
    public function batchProcess(&$context, Am_BatchProcessor $batch)
    {
        $start = $context;
        $pageCount = 10;
        $q = new Am_Query($this->getDi()->userTable);
        $count = 0;
        foreach ($q->selectPageRecords($start / $pageCount, $pageCount) as $u) {
            $count++;
            $this->_syncUser($u);
        }
        if (!$count) {
            return true;
        } else {
            $context += $count;
        }
    }

    public function delUser(Am_Event $e)
    {
        $user = $e->getUser();
        if ($id = $user->data()->get(self::DATA_KEY)) {
            $req = new Am_HttpRequest(self::API_ENDPOINT . '/users/' . $id, Am_HttpRequest::METHOD_DELETE);
            $req->setAuth($this->getConfig('pat'), '');
            $req->setHeader('Accept', 'application/json');
            $resp = $req->send();
            $this->log($req, $resp, sprintf('DELETE USER (%s)', $user->login));
        }
    }

    public function syncUser(Am_Event $e)
    {
        $this->_syncUser($e->getUser());
    }

    public function _syncUser(User $user)
    {
        $req = new Am_HttpRequest(self::API_ENDPOINT . '/users', Am_HttpRequest::METHOD_POST);
        $req->setAuth($this->getConfig('pat'), '');
        $req->setHeader('Accept', 'application/json');
        $req->setHeader('Content-Type', 'application/json');
        $cust = [
            'email' => $user->email,
            'user_id' => $user->pk(),
            'signed_up_at' => (int)amstrtotime($user->added),
            'name' => $user->getName(),
            'custom_attributes' => array(
                'payment_method' => $this->getPayment($user),
                'subscription' => $this->getSubscription($user)
            )
        ];
        $req->setBody(json_encode($cust));
        $resp = $req->send();
        $this->log($req, $resp, sprintf('SYNC USER (%s)', $user->login));
        if ($resp->getStatus() == '200') {
            $r = json_decode($resp->getBody(), true);
            $user->data()->set(self::DATA_KEY, $r['id'])
                ->update();
        }
    }

    function getPayment(User $user)
    {
        $pid = $this->getDi()->db->selectCell("SELECT paysys_id FROM ?_invoice_payment "
            . "WHERE user_id=? ORDER BY dattm DESC LIMIT 1", $user->pk());
        return $pid ?: 'free';
    }

    function getSubscription(User $user)
    {
        return implode(', ', $this->getDi()->productTable->getProductTitles($user->getActiveProductIds()));
    }

    function getWidget()
    {
        return $this->getDi()->auth->getUserId() ?
            ($this->getConfig('w_logged_in') ? $this->_getWidgetLoggedIn() : '') :
            ($this->getConfig('w_not_logged_in') ? $this->_getWidgetLoggedOut() : '');
    }

    function _getWidgetLoggedIn()
    {
        $user = $this->getDi()->auth->getUser();
        $s = [
            'app_id' => $this->getConfig('app_id'),
            'name' => $user->getName(),
            'email' => $user->email,
            'user_id' => $user->pk(),
            'created_at' => (int)amstrtotime($user->added),
            'subscription' => $this->getSubscription($user),
            'payment_method' => $this->getPayment($user)
        ];
        if ($this->getConfig('secure')) {
            $s['user_hash'] = hash_hmac('sha256', $user->email, $this->getConfig('secret'));
        }
        $s = json_encode($s);
        return <<<CUT
<script>
  window.intercomSettings = $s;
</script>
CUT
        . $this->_getCommonCode();
    }

    function _getWidgetLoggedOut()
    {
        $app_id = json_encode($this->getConfig('app_id'));
        return <<<CUT
<script>
  window.intercomSettings = {
    app_id: $app_id
  };
</script>
CUT
        . $this->_getCommonCode();
    }

    function _getCommonCode()
    {
        $app_id = json_encode($this->getConfig('app_id'));
        return <<<CUT
<script>(function(){var w=window;var ic=w.Intercom;if(typeof ic==="function"){ic('reattach_activator');ic('update',intercomSettings);}else{var d=document;var i=function(){i.c(arguments)};i.q=[];i.c=function(args){i.q.push(args)};w.Intercom=i;function l(){var s=d.createElement('script');s.type='text/javascript';s.async=true;s.src='https://widget.intercom.io/widget/$app_id';var x=d.getElementsByTagName('script')[0];x.parentNode.insertBefore(s,x);}if(w.attachEvent){w.attachEvent('onload',l);}else{w.addEventListener('load',l,false);}}})()</script>
CUT;

    }

    function log($req, $resp, $title)
    {
        if (!$this->_isDebug)
            return;
        $l = $this->getDi()->invoiceLogRecord;
        $l->paysys_id = $this->getId();
        $l->title = $title;
        $l->add($req);
        $l->add($resp);
    }
}