<?php

/**
 * @am_plugin_api 6.0
*/
class Am_Plugin_TwoFactorDuosecurity extends Am_Plugin_TwoFactor
{
    const PLUGIN_STATUS = self::STATUS_BETA;
    const PLUGIN_COMM = self::COMM_COMMERCIAL;
    const PLUGIN_REVISION = '6.3.21';

    function preauth(Am_Record $user, $ip)
    {
        $id = $this->getData($user);
        $resp = $this->_sendRequest("/preauth", [
            'ipaddr' => $ip,
            'user_id' => $id
        ]);
        if ($resp->getStatus() == 200) {
            $r = json_decode($resp->getBody(), true);
            if ($r['response']['result'] == 'allow')
                return true;
            $resp = $this->_sendRequest("/auth", [
                'display_username' => $user->login,
                'device' => 'auto',
                'factor' => 'push',
                'ipaddr' => $ip,
                'user_id' => $id,
                'async' => 1
            ]);
            if ($resp->getStatus() == '200') {
                $r = json_decode($resp->getBody(), true);
                $this->getSession()->txnid = $r['response']['txid'];
            }
        }
        return false;
    }

    function isValid(Am_Record $user, Am_Mvc_Request $r)
    {
        return $this->authStatus($r->getParam('txnid'));
    }

    function _initTwoFactorForm($form, $user)
    {
        $id = $this->getId();
        $txnid = $this->getSession()->txnid;
        $url = json_encode($this->getDi()->url('misc/' . $this->getId(),  [
                'txnid' => $txnid
        ], false));
        $form->addHtml(null, ['class' => 'no-lable'])
            ->setHtml(<<<CUT
<div style="background:#ffffcf; color:#454430; margin: -0.5em -1em; padding: 0.5em 1em">We'll send you a push request. Please use your Mobile Device to accept it.</div>
<script type="text/javascript">
var duoSecurityCheck = setInterval(function(){
    jQuery.get($url, null, function(res){
        if (res) {
            clearInterval(duoSecurityCheck);
            jQuery('#$id').submit();
        }
    })
}, 5000);
jQuery("[name=_submit], [name=_cancel]", "#$id").click(function(){
    clearInterval(duoSecurityCheck);
});
</script>
CUT
        );
        $form->addHidden('txnid')
            ->setValue($txnid);
    }

    function _initSetupForm(Am_Form_Setup $form)
    {
        parent::_initSetupForm($form);
        $form->setTitle(___('2Factor (Duosecurity.com)'));
        $form->addText('int_key', ['class' => 'am-el-wide'])
            ->setLabel(___('Integration key'));
        $form->addPassword('secret_key', ['class' => 'am-el-wide'])
            ->setLabel(___('Secret key'));
        $form->addText('host', ['class' => 'am-el-wide'])
            ->setLabel(___('API hostname'));
    }

    function isConfigured()
    {
        return $this->getConfig('int_key') &&
            $this->getConfig('secret_key') &&
            $this->getConfig('host');
    }

    function getEndpoint()
    {
        return sprintf('https://%s/auth/v2', $this->getConfig('host'));
    }

    function directAction(Am_Mvc_Request $request, Am_Mvc_Response $response, array $invokeArgs)
    {
        if ($id = $request->getParam('txnid')) {
            $response->ajaxResponse($this->authStatus($id));
        }
    }

    function authStatus($txnid)
    {
        $resp = $this->_sendRequest("/auth_status", [
            'txid' => $txnid
        ], 'GET');
        if ($resp->getStatus() == '200') {
            $r = json_decode($resp->getBody(), true);
            return ($r['response']['result'] == 'allow');
        }
        return false;
    }

    function checkStatus(Am_Record $user)
    {
        if ($this->isEnabled($user))
            return true;

        if ($id = $this->getData($user)) {
            $activation_code = $this->getData($user, 'activation_code');
            $resp = $this->_sendRequest('/enroll_status', [
                'user_id' => $id,
                'activation_code' => $activation_code
            ]);
            if ($resp->getStatus() == '200' &&
                ($r = json_decode($resp->getBody(), true)) &&
                $r["response"] == 'success') {
                $this->setData($user, 'activation_code', null);
                $this->setData($user, 'activation_barcode', null);
                $this->setData($user, 'expiration', null);
                $this->setData($user, 'enabled', 1);
                $user->save();
                $resp = $this->_sendRequest("/preauth", [
                    'user_id' => $id
                ]);
                if ($resp->getStatus() == '200') {
                    $r = json_decode($resp->getBody(), true);
                    list($device) = $r['response']['devices'];
                    $this->setData($user, 'device', $device['device']);
                    $this->setData($user, 'display_name', $device['display_name']);
                    $user->save();
                }
                return true;
            }
        }
        return false;
    }

    function register(Am_Record $user)
    {
        if (($e = $this->getData($user, 'expiration')) && $e > time()) {
            return $this->getData($user, 'activation_barcode');
        }
        $resp = $this->_sendRequest('/enroll', [
            'username' => get_class($user).'|'.$user->email,
        ]);

        if ($resp->getStatus() == 200) {
            $r = json_decode($resp->getBody(), true);
            $this->setData($user, null, $r['response']['user_id']);
            $this->setData($user, 'activation_code', $r['response']['activation_code']);
            $this->setData($user, 'activation_barcode', $r['response']['activation_barcode']);
            $this->setData($user, 'expiration', $r['response']['expiration']);
            $user->save();
            return $r['response']['activation_barcode'];
        } else {
            return false;
        }
    }

    function isPartiallyEnabled($user)
    {
        return (($e = $this->getData($user, 'expiration')) && $e > time());
    }

    protected function _sendRequest($endpoint, $params, $method = 'POST')
    {
        $req = new Am_HttpRequest(
                $this->getEndpoint() . $endpoint .
                (in_array($method, ['DELETE', 'GET']) && $params ? '?' . http_build_query($params) : ''),
                $method);
        $req->setHeader('Date', $date = date('r'));

        ksort($params);

        $hash = $date . "\n" .
            $method . "\n" .
            $this->getConfig('host') . "\n" .
            '/auth/v2' . $endpoint . "\n" .
            http_build_query($params);

        $hash = hash_hmac("sha1", $hash, $this->getConfig('secret_key'));

        if (!in_array($method, ['DELETE', 'GET'])) {
            $req->addPostParameter($params);
        }
        $req->setAuth($this->getConfig('int_key'), $hash);

        $resp = $req->send();
        $this->log($req, $resp, strtoupper(trim($endpoint, '/')));
        return $resp;
    }

    function getReadme()
    {
        $readme = parent::getReadme();
        return <<<CUT
You need to create <strong>Auth API</strong> Application in your duosecurity account
<a href="https://duo.com/docs/authapi" class="link" target="_blank" rel="noreferrer">https://duo.com/docs/authapi</a>

and fill in access credentials in from above.

{$readme}
CUT;
    }

}

abstract class TwoFactorDuosecurityController_Abstract extends Am_Mvc_Controller
{
    protected $layout;

    abstract function getAuth();
    abstract function _url($action = null);

    function getUser()
    {
        return $this->getAuth()->getUser();
    }

    function indexAction()
    {
        $user = $this->getUser();

        $this->view->title = ___('2Factor Authentication');
        if ($this->getPlugin()->isEnabled($user)) {
            $this->_disable($user);
        } else {
            $this->_enable($user);
        }
    }

    function checkStatusAction()
    {
        $user = $this->getUser();
        $this->_response->ajaxResponse($this->getPlugin()->checkStatus($user));
    }

    function _enable($user)
    {
        if ($this->getPlugin()->isPartiallyEnabled($user)) {
            return $this->_activation($user);
        }

        $form = new Am_Form();
        $form->addCsrf();
        $form->addSaveButton(___('Enable'));
        if ($form->isSubmitted() &&
            $form->validate()) {
            if ($id = $this->getPlugin()->getData($user)) {
                $this->getPlugin()->enable($user);
                $this->getAuth()->setUser($user);
                Am_Mvc_Response::redirectLocation($this->_url());
            } else {
                $this->_activation($user);
            }
        } else {
            $this->view->content = $form;
            $this->view->display($this->layout);
        }
    }

    function _activation($user)
    {
        if ($url = $this->getPlugin()->register($user)) {
            $status_url = $this->_url('check-status');
            $title = $this->getDi()->config->get('site_title');
            $this->view->content = <<<CUT
<h3>Activate Duo Mobile</h3>
<p>The $title Admin Panel is secured with two-factor authentication.
You'll use your phone to confirm your identity after you enter your username and password.</p>
<p>If you have an iPhone or Android device, the best way to authenticate is with Duo Mobile.</p>
<p>Launch the app store on your phone and search for "Duo Mobile." Install the app. Tap "OK" if asked about push notifications.
Open Duo Mobile and tap "Add Account." Then tap "Scan Barcode" and point your phone at this barcode:</p>
<img src="$url" />
<script type="text/javascript">
setInterval(function(){
    jQuery.get('$status_url', null, function(res){
        if (res)
           window.location.href = window.location.href;
    })
}, 5000);
</script>
CUT;
        } else {
            $this->view->content = <<<CUT
<ul class="am-errors">
    <li>Error occur. Please try to enroll later.</li>
</ul>
CUT;
        }
        $this->view->display($this->layout);
    }

    function _disable($user)
    {
        $form = new Am_Form();
        $form->addCsrf();
        $form->addStatic()
            ->setLabel(___('Status'))
            ->setContent(___('Enabled'));
        $form->addStatic()
            ->setLabel(___('Device'))
            ->setContent(($d = $this->getPlugin()->getData($user, 'display_name')) ? $d : ___('Unknown'));
        $form->addSaveButton(___('Disable'));
        if ($form->isSubmitted() &&
            $form->validate()) {

            $this->getPlugin()->disable($user);
            Am_Mvc_Response::redirectLocation($this->_url());
        } else {
            $this->view->content = $form;
            $this->view->display($this->layout);
        }
    }

    function getPlugin()
    {
        return $this->getDi()->plugins_misc->loadGet('two-factor-duosecurity');
    }
}

class TwoFactorDuosecurityController extends TwoFactorDuosecurityController_Abstract
{
    protected $layout = 'member/layout.phtml';

    function preDispatch()
    {
        $this->getDi()->auth->requireLogin();
    }

    function getAuth()
    {
        return $this->getDi()->auth;
    }

    function _url($action = null)
    {
        return $this->getDi()->url('two-factor-duosecurity' . ($action ? "/{$action}" : ""), false);
    }
}

class AdminTwoFactorDuosecurityController extends TwoFactorDuosecurityController_Abstract
{
    protected $layout = 'admin/layout.phtml';

    function checkAdminPermissions(Admin $admin)
    {
        return true;
    }

    function getAuth()
    {
        return $this->getDi()->authAdmin;
    }

    function _url($action = null)
    {
        return $this->getDi()->url('admin-two-factor-duosecurity' . ($action ? "/{$action}" : ""), false);
    }
}