<?php

class Am_Paysystem_PaypalCheckout extends Am_Paysystem_Abstract
{
    const
        PLUGIN_STATUS = self::STATUS_DEV,
        PLUGIN_DATE = '$Date$',
        PLUGIN_REVISION = '6.3.21',
        PAYPAL_CHECKOUT_BILLING_PLAN_ID = 'paypal-checkout-billing-plan-id',
        LIVE_API_URL = 'https://api-m.paypal.com/v1/',
        SANDBOX_API_URL = 'https://api-m.sandbox.paypal.com/v1/',
        SUBSCRIPTION_ID = 'paypal-checkout-subscription-id'
        ;

    protected
        $defaultTitle = "Paypal Checkout",
        $defaultDescription = "secure credit card payment";

    public function _initSetupForm(Am_Form_Setup $form)
    {
        $form->addText('client_id', ['class' => 'am-el-wide'])
            ->setLabel('Client ID')
            ->addRule('required');

        $form->addSecretText('secret', ['class' => 'am-el-wide'])
            ->setLabel("Secret")
            ->addRule('required');

        $form->addText('webhook_id', ['class' => 'am-el-wide'])
            ->setLabel('Webhook ID')
            ->addRule('required');

        $form->addAdvCheckbox("testing")
            ->setLabel("Is it a Sandbox(Testing) Account?");
    }

    function getToken()
    {
        $url = ($this->getConfig('testing') ? self::SANDBOX_API_URL : self::LIVE_API_URL) . 'oauth2/token';
        if(!($token = $this->getDi()->store->get($url)))
        {
            $req = new Am_HttpRequest($url, Am_HttpRequest::METHOD_POST);
            $req->setAuth($this->getConfig('client_id'), $this->getConfig('secret'))
                ->setHeader('Content-Type', 'application/x-www-form-urlencoded')
                ->setBody('grant_type=client_credentials');
            $res = $req->send();
            if(($vars = json_decode($res->getBody(), true)) && ($token = $vars['access_token']))
            {
                $this->getDi()->store->set($url, $token, "+" . $vars['expires_in'] . " seconds");
            }
            else
                return false;
        }
        return $token;
    }

    public function init()
    {
        parent::init();
        $this->getDi()->billingPlanTable->customFields()
            ->add(new Am_CustomFieldText(self::PAYPAL_CHECKOUT_BILLING_PLAN_ID, "Paypal Billing Plan ID",
                "you have to create billing plan using the same durations(prices could be different)"));
    }

    function isConfigured()
    {
        return strlen($this->getConfig('client_id')) &&
            strlen($this->getConfig('secret')) &&
            strlen($this->getConfig('webhook_id'));
    }

    function getSupportedCurrencies()
    {
        return array_keys(Am_Currency::getFullList());
    }

    function getRecurringType()
    {
        return self::REPORTS_REBILL;
    }

    public function _process($invoice, $request, $result)
    {
        $a = new Am_Paysystem_Action_HtmlTemplate('pay.phtml');
        $a->invoice = $invoice;

        $user = $invoice->getUser();
        if ($invoice->second_period && ($plan_id = $invoice->getItem(0)->getBillingPlanData(self::PAYPAL_CHECKOUT_BILLING_PLAN_ID)))
            $a->form = <<<CUT
<div id="paypal-button-container"></div>
<script src="https://www.paypal.com/sdk/js?client-id={$this->getConfig('client_id')}&vault=true&intent=subscription" data-sdk-integration-source="button-factory"></script>
<script>
paypal.Buttons({
    style: {
        shape: 'rect',
        color: 'gold',
        layout: 'vertical',
        label: 'subscribe'
    },
    createSubscription: function(data, actions) {
        return actions.subscription.create({
          custom_id: '{$invoice->public_id}',
          plan_id: '$plan_id',
          plan: {
              billing_cycles: [
                  {
                      sequence: 1,
                      total_cycles: 1,
                      pricing_scheme: {
                        fixed_price: {
                            currency_code: '{$invoice->currency}',
                            value: '{$invoice->first_total}'
                        }
                      }
                  },
                  {
                      sequence: 2,
                      total_cycles: {$invoice->rebill_times},
                      pricing_scheme: {
                        fixed_price: {
                            currency_code: '{$invoice->currency}',
                            value: '{$invoice->second_total}'
                        }
                      }
                  }
              ]},
        subscriber: {
          email_address: '{$invoice->getEmail()}',
          name: {
              given_name: '{$user->name_f}',
              surname: '{$user->name_l}'
          }
        },
        application_context: {
            return_url: '{$this->getReturnUrl()}',
            cancel_url: '{$this->getCancelUrl()}'
        }
      });
    },
    onApprove: function(data, actions) {
      window.location.replace("{$this->getReturnUrl()}"); 
    }    
  }).render('#paypal-button-container');
</script>
CUT;
        else
            $a->form = <<<CUT
<script src="https://www.paypal.com/sdk/js?currency={$invoice->currency}&client-id={$this->getConfig('client_id')}&components=messages,buttons&enable-funding=venmo"></script>
<div id="paypal-button-container"></div>
<script type="text/javascript">
  paypal.Buttons({
    createOrder: function(data, actions) {
      // This function sets up the details of the transaction, including the amount and line item details.
      return actions.order.create({
        purchase_units: [{
          reference_id: '{$invoice->public_id}',
          amount: {
            currency_code: '{$invoice->currency}',
            value: '{$invoice->first_total}'
          },
          description: '{$invoice->getLineDescription()}',
        }],
        payer: {
          email_address: '{$invoice->getEmail()}',
          name: {
              given_name: '{$user->name_f}',
              surname: '{$user->name_l}'
          }
        },
        application_context: {
            return_url: '{$this->getReturnUrl()}',
            cancel_url: '{$this->getCancelUrl()}'
        }
      });
    },
    onApprove: function(data, actions) {
      return actions.order.capture().then(function(details) {
        window.location.replace("{$this->getReturnUrl()}"); 
      });
    }    
  }).render('#paypal-button-container');
</script>
CUT;
        $result->setAction($a);
    }

    public function createTransaction($request, $response, array $invokeArgs)
    {
        return new Am_Paysystem_Transaction_PaypalCheckout($this, $request, $response, $invokeArgs);
    }

    function cancelAction(Invoice $invoice, $actionName, Am_Paysystem_Result $result) {
        $log = Am_Di::getInstance()->invoiceLogRecord;
        $req = new Am_HttpRequest(($this->getConfig('testing') ? self::SANDBOX_API_URL : self::LIVE_API_URL) .
            'billing/subscriptions/'. $invoice->data()->get(self::SUBSCRIPTION_ID) .'/cancel',
            Am_HttpRequest::METHOD_POST);
        $req->setHeader('Content-Type', 'application/json')
            ->setHeader('Authorization', 'Bearer ' . $this->getToken())
            ->setBody(json_encode([
                'reason' => 'Not satisfied with the service'
            ]));
        $res = $req->send();
        $log->setInvoice($invoice);
        $log->paysys_id = $invoice->paysys_id;
        $log->add($req);
        $log->add($res);
        if($res->getStatus() == 204) {
            $invoice->setCancelled(true);
            $result->setSuccess();
        }
        else
            throw new Am_Exception_InternalError('Paypal API error: Unable to cancel subscription');
    }

    public function getReadme() {
        $url = $this->getPluginUrl('ipn');
        return <<<CUT
In your Paypal account set the following url to listen Webhook for events: 
<strong>BILLING.SUBSCRIPTION.ACTIVATED, PAYMENT.SALE.COMPLETED, CHECKOUT.ORDER.APPROVED</strong>
 
Webhook Endpoint: <strong>$url</strong>

CUT;
    }

}

class Am_Paysystem_Transaction_PaypalCheckout extends Am_Paysystem_Transaction_Incoming
{
    protected $req = array();

    function __construct($plugin, $request, $response, $invokeArgs)
    {
        parent::__construct($plugin, $request, $response, $invokeArgs);
        $this->req = json_decode($request->getRawBody(), true);
    }

    public function validateSource()
    {
        if(openssl_verify(
                implode('|',
                    [
                    $this->request->getHeader('PAYPAL-TRANSMISSION-ID'),
                    $this->request->getHeader('PAYPAL-TRANSMISSION-TIME'),
                    $this->plugin->getConfig('webhook_id'),
                    crc32($this->request->getRawBody()),
                ]),
                base64_decode($this->request->getHeader('PAYPAL-TRANSMISSION-SIG')),
                openssl_pkey_get_public(file_get_contents( $this->request->getHeader('PAYPAL-CERT-URL'))),
                'sha256WithRSAEncryption'
            ) === 1)
        return true;

        //try to validate by sending webhook validation request
        if(!($token = $this->plugin->getToken()))
            return false;
        $req = new Am_HttpRequest(($this->plugin->getConfig('testing')
            ? Am_Paysystem_PaypalCheckout::SANDBOX_API_URL : Am_Paysystem_PaypalCheckout::LIVE_API_URL) .
            'notifications/verify-webhook-signature',
        Am_HttpRequest::METHOD_POST);
        $req->setHeader('Content-Type', 'application/json')
            ->setHeader('Authorization', 'Bearer ' . $token)
            ->setBody(json_encode([
                'auth_algo' => $this->request->getHeader('PAYPAL-AUTH-ALGO'),
                'cert_url' => $this->request->getHeader('PAYPAL-CERT-URL'),
                'transmission_id' => $this->request->getHeader('PAYPAL-TRANSMISSION-ID'),
                'transmission_sig' => $this->request->getHeader('PAYPAL-TRANSMISSION-SIG'),
                'transmission_time' => $this->request->getHeader('PAYPAL-TRANSMISSION-TIME'),
                'webhook_id' => $this->plugin->getConfig('webhook_id'),
                'webhook_event' => $this->req
            ]));
        $res = $req->send();
        $this->log->add($req);
        $this->log->add($res);
        if($res->getStatus() != 200) return false;
        $_ = json_decode($res->getBody(), true);
        if(empty($_['verification_status'])) return false;
        return $_['verification_status'] == 'SUCCESS';
    }

    public function validateTerms()
    {
        return true;
    }

    public function validateStatus()
    {
        return true;
    }

    function getUniqId()
    {
        return $this->req['resource']['id'];
    }

    function findInvoiceId()
    {
        return !empty($this->req['resource']['custom_id']) ?
            $this->req['resource']['custom_id'] :
            (!empty($this->req['resource']['custom']) ?
            $this->req['resource']['custom'] :
            $this->req['resource']['purchase_units'][0]['reference_id']);
    }

    function processValidated()
    {
        switch ($this->req['event_type'])
        {
            case 'BILLING.SUBSCRIPTION.ACTIVATED':
                $this->invoice->data()->set(Am_Paysystem_PaypalCheckout::SUBSCRIPTION_ID, $this->req['resource']['id'])->update();
                break;
            case 'PAYMENT.SALE.COMPLETED':
            case 'CHECKOUT.ORDER.APPROVED':
                $this->invoice->addPayment($this);
                break;
        }
    }
}