import React, {FC, useCallback, useEffect, useState} from "react";
import {CategoryFormContainer, ErrorMessage, FormRow, InputLabel, SaveButton, WideInput} from "./Styled";
import {TCategory, TCategoryForm, TCategorySelectOption, THtmlEditor} from "../types";
// @ts-ignore
import {CKEditor} from '@ckeditor/ckeditor5-react';
// @ts-ignore
import ClassicEditor from '@ckeditor/ckeditor5-build-classic';
import Loading from "../../Utils/Loading";
import {observer} from "mobx-react";
import {useCategoryStore} from "../Store";
import Select, {SingleValue} from "react-select";
import {useDropzone} from "react-dropzone";
import Img from "../../Marketplace/Layout/Listing/Img";
import noop from "../../Utils/noop";

declare function amUrl(url: string):string

const HtmlEditor: THtmlEditor = (props) => {

    const onChange = props.onChange ?? (() => {
    })

    return <CKEditor
        editor={ClassicEditor}
        data={props.value}
        onReady={(editor: any) => {
        }}
        onChange={(event: any, editor: any) => {
            const data = editor.getData();
            onChange(data);
        }}
        onBlur={(event: any, editor: any) => {
        }}
        onFocus={(event: any, editor: any) => {
        }}
    />
}
const ImageUpload: FC<{maxUploadWidth: number, maxUploadSize: number, onSetFile: (file: File) => void, onDeleteFile: () => void, url?: string}> = (props) => {
    const [preview, setPreview] = useState<string>();
    const {acceptedFiles, getRootProps, getInputProps} = useDropzone({
        maxSize: props.maxUploadSize * 1024,
        onDrop: (acceptedFiles, fileRejections, event) => {
            if (acceptedFiles[0] !== undefined) {
                props.onSetFile(acceptedFiles[0]);
                setPreview(URL.createObjectURL(acceptedFiles[0]));
            }
        }
    });

    useEffect(() => {
        setPreview(props.url);
    }, [props.url]);
    return (
        <div  {...getRootProps({ className: 'dropzone', style: {"border":"dashed 1px", padding: "10px"}})}>
            {preview && <>
                <div><Img src={preview} width={props.maxUploadWidth}/></div>
                <div><a href={"#"} onClick={(event) => {props.onDeleteFile(); event.stopPropagation(); event.preventDefault();}}>Delete</a></div>

            </>}
                <input {...getInputProps()} />
                <p>Attach an image to listing</p>
                <p>Drag 'n' drop file here, or click to select
                    file</p>
                <p >One image file accepted(jpg, png, gif).
                    Max {props.maxUploadSize}Kb</p>
        </div>
    );

}


const CategoryForm: TCategoryForm = observer((props) => {
    const store = useCategoryStore()
    const [category, setCategory] = useState<TCategory | undefined>();
    const [formError, setFormError] = useState("")
    const [isLoading, setLoading] = useState(false)
    const[file, setFile] = useState<File>();

    const onSubmit = props.onSubmit ?? (() => Promise.resolve())


    useEffect(() => {
        setCategory(props.category);
    }, [props.category]);
    const updateCategory = (data: Record<any, any>) => {
        setCategory({...category, ...data})
    }

    const validate = () => {
        const error = !((category?.name ?? "").length > 0);
        return category && !error ? Promise.resolve(category) : Promise.reject();
    }

    const formSubmit = () => {
        setLoading(true)
        validate().then(
            (category) => store.apiSaveCategory(category)
                .then(() => {
                    setFormError("");
                    onSubmit(category);
                    store.closeForm();
                })
                .catch((e) => {
                    setFormError(e.message);
                })
        ).catch(() => {
            setFormError("Name can't be empty");
        }).finally(() => {
            setLoading(false);
        })

    }
    const uploadFile = async (file: File) =>
    {
        try {
            const data = new FormData()
            data.append('upload', file);

            const response = await fetch(store.uri + '/upload', {
                method: 'POST',
                mode: 'cors',
                cache: 'no-cache',
                credentials: 'same-origin',
                redirect: 'follow',
                referrerPolicy: 'no-referrer',
                body: data
            });

            const json = await response.json();

            if (json.ok === false)
                throw new Error(json.errorMessage);

            updateCategory({image: json.upload_id});

    } catch (error) {
            setFormError((error as Error).message);
    }

    }
    const customStyles = {
        option: (provided: any, state: any) => {
            console.log(state.value);
            return {
                ...provided,
                paddingLeft: ((store.getCategoryOptionById(state.value).indent ?? 0) * 2 + 1) + "em",
            }
        }
    }
    return (
        <>{props.show && <CategoryFormContainer>
            {formError && <FormRow>
                <ErrorMessage>{formError}</ErrorMessage>
            </FormRow>}

            <FormRow>
                <InputLabel>Category Name</InputLabel>
                <WideInput autoFocus name="categoryName" error={formError.length > 0} value={category?.name}
                           onChange={(ev) => updateCategory({name: ev.target.value})}/>
            </FormRow>
            <FormRow>
                <InputLabel>Description</InputLabel>
                <HtmlEditor show={true} value={category?.description}
                            onChange={(value) => updateCategory({description: value})}/>
            </FormRow>
            <FormRow>
                <InputLabel>Category Default Image</InputLabel>
                <ImageUpload maxUploadWidth={store.max_upload_width} maxUploadSize={store.max_upload_size} onDeleteFile={()=>{updateCategory({image: null})}} onSetFile={uploadFile}
                url={category?.image?amUrl('/upload/get/id/'+category.image) : ""}></ImageUpload>
            </FormRow>
            <FormRow>
                <InputLabel>Parent Category</InputLabel>
                <Select options={store.categorySelectOptions} value={store.getCategoryOptionById(category?.parent_id)}
                        styles={customStyles}
                        onChange={(option: SingleValue<{}>) => {
                            updateCategory({parent_id: (option as unknown as TCategorySelectOption).value})
                        }}/>

            </FormRow>
            <FormRow>
                <SaveButton disabled={isLoading} title={props.category?.category_id ? "Update" : "Create"}
                            onClick={formSubmit}/>
                <SaveButton title={"Close"} onClick={props.onClose}/>
                {isLoading && <Loading/>}
            </FormRow>
        </CategoryFormContainer>}
        </>
    );
})

export default CategoryForm;
