import {makeAutoObservable, runInAction} from "mobx";
import {createContext, useContext} from "react";
import {
    TChatConfig,
    TChatEvent,
    TChatEventListener,
    TChatEventType,
    TChatMessage,
    TChatMessagesCollection,
    TChatStoreState
} from "../types";

import MD5 from "crypto-js/md5";


import postData from "../../Utils/postData";
import {TApiResponse, TApiResponseList} from "../../types";

export class ChatStore {
    private config: TChatConfig
    messages: TChatMessagesCollection = []
    total: number | undefined
    error?: string
    state: TChatStoreState = "ok"
    listeners?: Record<TChatEventType, {
        callbacks: Array<TChatEventListener>
        signatures: Array<string>
    }>;
    events: Array<ChatEvent> = [];


    constructor(config: TChatConfig) {
        this.config = config;
        makeAutoObservable(this, {
            listeners: false,
            events: false
        });
    }

    setConfig(config: TChatConfig) {
        this.config = config;
    }


    get latestMessageId(): number | null {
        return this.messages[this.messages.length - 1]?.message_id;
    }

    get firstMessageId(): number | null {
        return this.messages[0].message_id;
    }

    async apiRequest(endpoint: string, data: {} = {}): Promise<TApiResponse> {

        this.state = 'loading';
        try {
            return await postData(this.apiUrl(endpoint), data).then((response) => {
                runInAction(() => {
                    this.state = 'ok';
                    this.error = undefined;
                })
                return response;
            })
        } catch (error) {
            runInAction(() => {
                this.setError((error as Error).message);
                this.state = 'ok';
            })
            return {ok: false, errorMessage: (error as Error).message, errorCode: 500}

        }
    }

    apiUrl(endpoint: string): string {
        return this.config.apiUrl + endpoint;
    }


    async apiGetMessages(): Promise<any> {
        if (this.state == 'loading')
            return Promise.resolve();

        return await this.apiRequest('/list', {
            last_id: this.latestMessageId,
            transaction_id: this.config.chatId
        }).then(response => this.updateList(response))
    }

    async apiGetHistory(): Promise<any> {
        return await this.apiRequest('/list', {
            last_id: this.firstMessageId,
            transaction_id: this.config.chatId,
            history: true
        }).then((response => this.updateList(response, true)))
    }

    updateList(response: TApiResponseList<TChatMessage>, history: boolean = false) {
        if (response.list.length > 0) {
            if (history) {
                this.messages = [...response.list, ...this.messages];
            } else {
                this.messages = [...this.messages, ...response.list];
            }
            this.total = response.total;
            //this.event("ChatUpdate");
        }
    }

    async apiPostMessage(message: string) {
        return await this.apiRequest('/post', {
            message: message,
            transaction_id: this.config.chatId,
            last_id: this.latestMessageId
        }).then(response => this.updateList(response))
    }

    setError(message: string) {
        this.error = message;
        this.state = 'error';
    }

    get ownerId() {
        return this.config.ownerId;
    }

    get hasMore() {
        return (this.total ?? 0) > this.messages.length;
    }

    private getSignature(callback: TChatEventListener) {
        return MD5(callback.toString()).toString();
    }

    /**
     addListener(type: TChatEventType, callback: TChatEventListener)
     {
        const  sig = this.getSignature(callback);

        if(!this.listeners || !Object.hasOwn(this.listeners, type))
        {
                this.listeners = Object.assign(this.listeners??{}, {[type]: {
                    callbacks : [],
                    signatures: []
                    }});
        }
        if(this.listeners[type].signatures.indexOf(sig)===-1)
        {
            this.listeners[type].signatures.push(sig);
            this.listeners[type].callbacks.push(callback);
        }
    }

     private event(type: TChatEventType, data?: Record<any, any>)
     {
        const event = new ChatEvent(type);
        if(this.listeners && Object.hasOwn(this.listeners, type)){
            this.listeners[type].callbacks.map(callback =>{
               callback(event);
            });
        }
    }
     */

}


class ChatEvent implements TChatEvent {
    data: Record<any, any> = {};
    type: TChatEventType;

    constructor(type: TChatEventType, data: Record<any, any> = {}) {
        this.type = type
        this.data = data;
    }

}

export const ChatStoreContext = createContext<ChatStore>({} as ChatStore);
export const useChatStore = () => useContext(ChatStoreContext);
export const useChatEvent = (type: TChatEventType, callback: TChatEventListener) => {
    const store = useChatStore();
    //store.addListener(type, callback);
}