import {makeAutoObservable, runInAction} from "mobx";
import {MarketplaceStore} from "./index";
import {TApiGetListingsOptions, TFilter, TListing, TListingPublicData} from "../types";
import {TApiResponseList} from "../../types";
import theme from "../theme";

export class ListingStatus {
    status: string = 'pending';

    constructor(status: string) {
        this.status = status;
    }

    get text() {
        switch (this.status) {
            case 'published':
                return "Published";
            case 'draft':
                return "Draft";
            case 'sold':
                return 'Sold';
            case 'hold' :
                return "Waiting For Safe Deal"
            default :
                return "Unknown";
        }
    }

    get color() {

        switch (this.status) {
            case 'published':
                return theme.palette.success.light
            case 'draft':
                return theme.palette.secondary.dark
            case 'sold':
                return theme.palette.info.light
            case 'hold' :
                return theme.palette.warning.light
            default :
                return theme.palette.secondary.main;
        }
    }

    get active() {
        return this.status == 'published';
    }

    get draft() {
        return this.status == 'draft';
    }

    get purchased() {
        return this.status == 'sold';
    }
}

export class ListingRecord implements TListingPublicData {

    record: TListing

    constructor(record: TListing) {
        this.record = record;
    }

    get title() {
        return this.record.title;
    }

    get categories() {
        return this.record.categories ?? [];
    }

    get description() {
        return this.record.description ?? "";
    }

    get price() {
        return this.record.price ?? "";
    }

    get owner() {
        return {
            username: this.record.owner.username,
            totalSold: this.record.owner.total ?? 0,
            rating: this.record.owner.rating ?? 0
        };
    }

    get id() {
        return this.record.listing_id
    }

    get created() {
        return this.record.created
    }

    get status() {
        return new ListingStatus(this.record.status);
    }

    get approved() {
        return this.record.is_approved ?? false;
    }

    get image() {
        return this.record.imageUrl ?? null;
    }

    get qty() {
        return this.record.qty ?? 1;
    }

    get views() {
        return this.record.views ?? 0;
    }

    get delivery() {
        return this.record.delivery ?? 0;
    }

    get deliveryText() {
        return this.listingType == 'regular' ? this.delivery + ' hours' : "Immediate delivery";
    }

    get listingType() {
        return this.record.listing_type ?? 'regular';
    }

    get listingTypeText() {
        return this.listingType == 'regular' ? "Deliverable Item" : "Serial Code"
    }
}


export class PurchasedListingRecord extends ListingRecord {
    get instructions(): string {
        return this.record.instructions ?? ""
    }
}

export default class ListingStore {
    rootStore: MarketplaceStore
    sort: string = 'created-desc'
    count: number = 20
    totalRecords: number = 0;
    skipFirst: number = 0;
    list: Array<ListingRecord> = [];
    singleListing: ListingRecord = {} as ListingRecord
    sortOptions = [
        {
            id: "created-desc",
            title: "Newest First"
        },
        {
            id: "created",
            title: "Oldest First"
        },
        {
            id: "price",
            title: "Price: Low to High"
        },
        {
            id: "price-desc",
            title: "Price: Hight to Low"
        },
        {
            id: "rating",
            title: "User Rating"
        },
        {
            id: "delivery",
            title: "Fastest Delivery"
        }

    ]

    constructor(rootStore: MarketplaceStore) {
        this.rootStore = rootStore;

        makeAutoObservable(this, {
            rootStore: false,
        });
    }


    async apiGetListings(options: TApiGetListingsOptions) {
        const response = await this.rootStore.apiRequest('/listing/list' + (options.all ? "" : "/my"), {
            filter: options.filter ?? {},
            sort: this.sort,
            count: this.count,
            skipFirst: options.skipFirst ?? 0
        });
        if (response.ok == true) {
            runInAction(() => {
                    let r = response as TApiResponseList<TListing>;
                    this.totalRecords = r.total
                    let list = r.list.map<ListingRecord>((el) => new ListingRecord(el));
                    this.list = options.skipFirst !== undefined ? [...this.list, ...list] : list;
                }
            )

        }
        return response as TApiResponseList<TListing>;
    }

    async apiGetSingleListing(id: string, type: string) {
        const response = await this.rootStore.apiRequest('/listing/list' + (type == "all" ? "" : "/my"), {
            id: id
        }) as TApiResponseList<TListing>;
        const listing = response.list.pop();
        if (listing !== undefined) {
            return new ListingRecord(listing);
        }
        return Promise.reject("Not found");
    }


    getFilter(params: URLSearchParams): TFilter {
        const filter = {} as TFilter;
        filter.text = params.get('text') ?? "";
        this.rootStore.categories.types.map((type) => {
            filter.category = Object.assign({}, filter.category ?? {}, {[type.type_id]: params.get('cat-' + type.slug)})
        })
        return filter;
    }


    setSort(sort: string | undefined) {
        this.sort = sort ?? "date-desc";
    }

    get sortTitle() {
        return this.sortOptions.reduce((prev, curr) => (curr.id == this.sort ? curr.title : prev), "");
    }

    getListing(id: string) {
        return this.list.reduce<ListingRecord | null>((prev, current) => (current.id == id ? current : prev), null);
    }

}

