<?php
/**
 * @property int $seller_id
 * @property int $buyer_id
 * @property $started
 * @property $finished
 * @property string $status
 * @property string $comment
 * @property int $qty
 * @property int $price
 * @property int $total
 * @property string $listing
 * @property string $title
 *
 */

class MarketplaceTransaction extends Am_Record_WithData implements JsonSerializable
{
    const
        STATUS_PENDING = 'pending',
        STATUS_COMPLETED = 'completed',
        STATUS_REVERTED = 'reverted',
        STATUS_EXPIRED = 'expired',
        STATUS_SENT = 'sent';
    
    
    /**
     * @var MarketplaceListing $_listing
     */
    protected $_listing;
    /**
     * @var User
     */
    protected $_buyer;
    
    /**
     * @var User
     */
    protected $_seller;
    
    static function statusText($status)
    {
        switch ($status) {
            case self::STATUS_PENDING :
                return ___('Pending');
            case self::STATUS_REVERTED :
                return ___('Reverted');
            case self::STATUS_COMPLETED :
                return ___('Completed');
            case self::STATUS_EXPIRED :
                return ___('Expired');
            case self::STATUS_SENT :
                return ___('Sent');
        }
    }
    
    function setListing(MarketplaceListing $listing, $qty = 1): self
    {
        $this->_listing = $listing;
        $this->listing = json_encode($listing->getTransactionInfo());
        $this->title = $listing->title;
        $this->qty = $qty;
        $this->setSeller($listing->getUser());
        $this->price = $listing->price;
        $this->total = $this->qty * $this->price;
        $deliveryDefault = $this->getDi()->modules->loadGet('marketplace')->getConfig('seller-delivery-timeframe');
        $this->expire = $this->getDi()->dateTime
            ->modify("+" . (intval($listing->delivery) > 0 ? $listing->delivery : $deliveryDefault) . ' hours')
            ->format("Y-m-d H:i:s");
        
        return $this;
    }
    
    function getListing()
    {
        if ($this->_listing) {
            return $this->_listing;
        }
        try {
            $listing = json_decode($this->listing, true);
            
            if (empty($listing['listing_id'])) {
                return null;
            }
            
            $this->_listing = $this->getDi()->marketplaceListingTable->load($listing['listing_id']);
            return $this->_listing;
        } catch (Exception $ex) {
            return null;
        }
    }
    
    function setSeller(User $seller): self
    {
        $this->_seller = $seller;
        $this->seller_id = $seller->pk();
        return $this;
    }
    
    function getSeller(): User
    {
        if (empty($this->_seller)) {
            $this->_seller = $this->getDi()->userTable->load($this->seller_id);
        }
        return $this->_seller;
    }
    
    function setBuyer(User $buyer): self
    {
        $this->_buyer = $buyer;
        $this->buyer_id = $buyer->pk();
        return $this;
    }
    
    function getBuyer(): User
    {
        if (empty($this->_buyer)) {
            $this->_buyer = $this->getDi()->userTable->load($this->buyer_id);
        }
        return $this->_buyer;
    }
    
    function recordTransactionUpdate(User $user, $delete = false): self
    {
        $data = $user->data()->getBlob("marketplace-notify");
        $type = $this->buyer_id == $user->pk() ? "orders" : "sales";
        if (!empty($data)) {
            $data = json_decode($data, true);
        } else {
            $data = ['sales' => [], 'orders' => []];
        }
        $data[$type][] = $this->pk();
        $callback = $delete ? function ($item)
        {
            return $item != $this->pk();
        } : function ($item)
        {
            return !empty($item);
        };
        $data[$type] = array_filter(array_unique($data[$type]), $callback);
        
        $user->data()->setBlob("marketplace-notify", json_encode($data));
        $user->data()->update();
        $this->updated = $this->getDi()->sqlDateTime;
        $this->update();
        
        return $this;
    }
    
    function complete()
    {
        $this->status = self::STATUS_COMPLETED;
        $this->finished = $this->getDi()->sqlDateTime;
        $this->updated = $this->getDi()->sqlDateTime;
        /**
         * @var Bootstrap_Marketplace $module
         */
        $module = $this->getDi()->modules->get('marketplace');
        
        $listing = $this->getListing();
        
        if (!empty($listing)) {
            if ($listing->listing_type == MarketplaceListing::TYPE_SERIAL) {
                $listingSerials = $listing->getSerialCodes();
                $codes = array_splice($listingSerials, 0, $this->qty);
                $this->serial_codes = json_encode($codes);
                $listing->setSerialCodes($listingSerials);
                if (empty($listingSerials)) {
                    $listing->qty = 0;
                    $listing->status = MarketplaceListing::STATUS_HOLD;
                } else {
                    $listing->qty = count($listingSerials);
                }
                
            }
            if ($listing->status == MarketplaceListing::STATUS_HOLD) {
                $listing->status = MarketplaceListing::STATUS_SOLD;
            }
            
            $listing->update();
        }
        
        $module->creditsHistoryTable()->credit($this->getSeller(), $this->total, 1,
            "Markeplace transaction# {$this->pk()} listing# {$this->title}", 'MarketplaceTransactionTable',
            $this->pk());
        $this->recordTransactionUpdate($this->getSeller());
        $this->update();
    }
    
    function setComment(string $comment)
    {
        $this->comment = $comment;
        $this->update();
        $this->recordTransactionUpdate($this->getSeller());
    }
    
    function markAsSent()
    {
        $this->status = self::STATUS_SENT;
        $this->update();
        $this->recordTransactionUpdate($this->getBuyer());
    }
    
    function revert()
    {
        $this->status = self::STATUS_REVERTED;
        $this->updated = $this->getDi()->sqlDateTime;
        /**
         * @var Bootstrap_Marketplace $module
         */
        $module = $this->getDi()->modules->get('marketplace');
        
        $listing = $this->getListing();
        if (!empty($listing)) {
            if ($listing->status == MarketplaceListing::STATUS_HOLD) {
                $listing->status = MarketplaceListing::STATUS_PUBLISHED;
            }
            $listing->qty += $this->qty;
            $listing->update();
        }
        
        $module->creditsHistoryTable()->credit($this->getBuyer(), $this->total, 1,
            ___('Marketplace order revert# ' . $this->pk()),
            'marketplace_transaction', $this->pk());
        
        $this->recordTransactionUpdate($this->getBuyer());
        $this->update();
        
    }
    
    public function update()
    {
        
        $this->log(
            "Transaction was updated. Status: {$this->status}"
        );
        $this->updated = $this->getDi()->sqlDateTime;
        return parent::update(); // TODO: Change the autogenerated stub
    }
    
    function log(string $record)
    {
        /**
         * @var Bootstrap_Marketplace $module
         */
        $module = $this->getDi()->modules->get('marketplace');
        $log = $module->logTable()->createRecord();
        $log->record = $record;
        $log->user_id = $this->getDi()->auth->getUserId();
        $log->remote_addr = $this->getDi()->request->getClientIp();
        $log->user_agent = $this->getDi()->request->getServer('HTTP_USER_AGENT');
        $log->date = $this->getDi()->sqlDateTime;
        $log->transaction_id = $this->pk();
        $log->insert();
    }
    
    function insert($reload = true)
    {
        /**
         * Update listing
         * Update user credits
         */
        $this->getListing()->pendingTransaction($this);
        $ret = parent::insert();
        
        $this->log(
            "Transaction was created. Status: {$this->status}"
        );
        
        $this->recordTransactionUpdate($this->getSeller());
        return $ret;
        
    }
    
    function getBuyerLogin($user_id)
    {
        return $this->getDi()->cacheFunction->call([$this->getDi()->userTable, 'load'], [$user_id])->login;
    }
    
    public function jsonSerialize()
    {
        $data = $this->toArray();
        $data['listing'] = json_decode($data['listing']);
        $data['statusText'] = self::statusText($this->status);
        $data['buyer'] = $this->getBuyerLogin($this->buyer_id);
        $data['updated'] = amDatetime($data['updated'] ?? $data['started']);
        $data['started'] = amDatetime($data['started']);
        $data['finished'] = $data['finished'] ? amDatetime($data['finished']) : "";
        $data['expire'] = $data['expire'] ? amDatetime($data['expire']) : '';
        $data['serial_codes'] = json_decode($data['serial_codes']) ?? [];
        return $data;
    }
    
    /**
     * Chec k whenever user is transaction owner (buyer or seller)
     * @param User $user
     * @return bool
     */
    function isOwner(User $user)
    {
        return in_array($user->pk(), [$this->buyer_id, $this->seller_id]);
    }
    
    function chat(): MarketplaceTransactionChatTable
    {
        return $this->getDi()->marketplaceTransactionChatTable->setTransaction($this);
    }
    
}

class MarketplaceTransactionTable extends Am_Table_WithData
{
    
    protected $_table = '?_marketplace_transaction';
    protected $_key = 'transaction_id';
    protected $_recordClass = 'MarketplaceTransaction';
    
    public function getSellerRating(int $seller_id): string
    {
        $totalTransactions = $this->_db->selectCell("select count(*) from {$this->_table} where seller_id=?",
            $seller_id);
        return sprintf("%02d", ($this->getTotalSold($seller_id) / $totalTransactions) * 100);
    }
    
    public function getTotalSold(int $seller_id): int
    {
        return $this->_db->selectCell("
                select count(*)
                from {$this->_table}
                where seller_id =? and status=?",
            $seller_id,
            MarketplaceTransaction::STATUS_COMPLETED);
    }
    
    public function createTransaction()
    {
        /**
         * @var  MarketplaceTransaction $record
         */
        $record = $this->createRecord();
        $record->started = $this->getDi()->sqlDateTime;
        $record->updated = $this->getDi()->sqlDateTime;
        $record->status = MarketplaceTransaction::STATUS_PENDING;
        return $record;
        
        
    }
    
}