<?php
/**
 * @title Marketplace
 * @desc Simple marketplace to be used with self-service store
 * @setup_url marketplace/admin
 */

class Bootstrap_Marketplace extends Am_Module
{
    const PRODUCT_ENABLED = 'marketplace-enabled';
    const PRODUCT_ENABLE_QTY = 'marketplace-enable-qty';
    const PRODUCT_ACCESS = 'marketplace-access';
    const ADMIN_PERM = 'marketplace';
    const UPLOAD_PREFIX = 'marketplace-listing';
    
    function init()
    {
        $router = $this->getDi()->router;
        $router->addRoute('marketplace-index', new Am_Mvc_Router_Route_Regex(
            'marketplace/((?!admin|ajax).)*',
            [
                'module' => $this->getId(),
                'controller' => 'index',
                'action' => 'index',
            ],
        ));
        
        $router->addRoute('marketplace-ajax', new Am_Mvc_Router_Route(
            'marketplace/ajax/*',
            [
                'module' => $this->getId(),
                'controller' => 'ajax',
                'action' => 'index',
            ]
        ));
        
        $this->getDi()->uploadTable->defineUsage(self::UPLOAD_PREFIX,
            'marketplace_listing', 'image',
            UploadTable::STORE_FIELD, "Listing Image", '');
        
    }
    
    function onInitBlocks(Am_Event $event)
    {
        if($form_id = $this->getConfig("buy-credits-form")){
            $event->getBlocks()->add(
                "signup/form/before",
                new Am_Block_Base(
                    ___('Marketplace Item Receipt'),
                    '_marketplace_receipt',
                    null,
                    function(Am_View $view) use ($form_id)
                    {
                        if($form_id == $view->record->saved_form_id)
                        {
                            $data = $this->getPurchaseData();
                            if(!empty($data['listing_id'])){
                                $listing = $this->listingsTable()->load($data['listing_id']);
                                $qty = $data['qty']??1;
                                $user = $this->getDi()->auth->getUser();
                                if($listing->isPurchasePossible($user, $qty)){
                                    $view->assign('listing', $listing);
                                    $view->assign('qty', $qty);
                                    $view->assign('user', $user);
                                    $view->assign('balance', $this->getBalance($user));
                                    return $view->render("marketplace/blocks/signup.phtml");
                                }
                            }
                            
                        }
                        return null;
                    })
            );
            
        }
    }
    
    function getBalance(User $user): int
    {
        return $this->getDi()->selfserviceBalanceTable->balance($user) ?? 0;
    }
    
    
    function getPurchaseData()
    {
        return $this->getDi()->session->ns('marketplace')->purchaseData;
    }
    function setPurchaseData($data)
    {
        $ns = $this->getDi()->session->ns('marketplace');
        $ns->setExpirationSeconds(3600);
        $ns->purchaseData = $data;
    }
    
    function onGetUploadPrefixList(Am_Event $e)
    {
        $e->addReturn([
            Am_Upload_Acl::IDENTITY_TYPE_ADMIN => Am_Upload_Acl::ACCESS_ALL,
            Am_Upload_Acl::IDENTITY_TYPE_USER => Am_Upload_Acl::ACCESS_ALL,
            Am_Upload_Acl::IDENTITY_TYPE_ANONYMOUS => Am_Upload_Acl::ACCESS_ALL
        ], self::UPLOAD_PREFIX);
    }
    
    function onGetPermissionsList(Am_Event $e)
    {
        $e->addReturn("Marketplace", self::ADMIN_PERM);
    }
    
    function onAdminMenu(Am_Event $event)
    {
        $menu = $event->getMenu();
        $awaitApproval = $this->getDi()->db->selectCell("SELECT COUNT(*) FROM ?_marketplace_listing WHERE is_approved=0 and status=?",
            MarketplaceListing::STATUS_PUBLISHED);
        $menu->addPage([
            'id' => 'marketplace',
            'uri' => 'javascript:;',
            'label' => ___('Marketplace'),
            'order' => 80,
            'resource' => self::ADMIN_PERM,
            'pages' => [
                [
                    'id' => 'marketplace-categories',
                    'controller' => 'admin-categories',
                    'module' => 'marketplace',
                    'label' => ___('Listing Categories'),
                    'resource' => self::ADMIN_PERM
                ],
                [
                    'id' => 'marketplace-listings',
                    'controller' => 'admin-listings',
                    'module' => 'marketplace',
                    'action' => 'index',
                    'label' => ___('All Listings'),
                    'resource' => self::ADMIN_PERM
                ],
                [
                    'id' => 'marketplace-not-approved',
                    'controller' => 'admin-listings-na',
                    'module' => 'marketplace',
                    'action' => 'index',
                    'label' => ___('Not Approved') . ($awaitApproval ? " (($awaitApproval))" : ''),
                    'resource' => self::ADMIN_PERM
                ],
                [
                'id' => 'marketplace-transaction',
                'controller' => 'admin-transactions',
                'module' => 'marketplace',
                'action' => 'index',
                'label' => ___('Transactions'),
                'resource' => self::ADMIN_PERM
            ]
            
            
            ]
        
        ]);
    }
    
    function onGridProductInitForm(Am_Event_Grid $event)
    {
        $form = $event->getGrid()->getForm();
        $fs = $form->addFieldset('')->setLabel(___('Marketplace'));
        $fs->addAdvCheckbox('_marketplace_access')->setLabel(___('Allow to access marketplace'));
        $fs->addAdvCheckbox('_marketplace_enabled')->setLabel(___('Allow to post listings'));
        $fs->addAdvCheckbox('_marketplace_qty')->setLabel(___('Enable quantity in listings'));
    }
    
    function onGridProductInitGrid(Am_Event_Grid $event)
    {
        $event->getGrid()->addCallback(Am_Grid_Editable::CB_VALUES_FROM_FORM, function (&$values, Product $record)
        {
            $record->data()
                ->set(self::PRODUCT_ENABLED, $values['_marketplace_enabled'])
                ->set(self::PRODUCT_ENABLE_QTY, $values['_marketplace_qty'])
                ->set(self::PRODUCT_ACCESS, $values['_marketplace_access']);
        });
        $event->getGrid()->addCallback(Am_Grid_Editable::CB_VALUES_TO_FORM, function (&$values, $record)
        {
            $values['_marketplace_enabled'] = $record->data()->get(self::PRODUCT_ENABLED);
            $values['_marketplace_qty'] = $record->data()->get(self::PRODUCT_ENABLE_QTY);
            $values['_marketplace_access'] = $record->data()->get(self::PRODUCT_ACCESS);
        });
    }
    
    function onUserTabs(Am_Event_UserTabs $event)
    {
        if ($event->getUserId() > 0) {
            $user = $this->getDi()->userTable->load($event->getUserId());
            $event->getTabs()->addPage([
                'id' => 'marketplace',
                'uri' => '#',
                'label' => ___('Marketplace'),
                'order' => 1000,
                'resource' => self::ADMIN_PERM,
                'pages' => [
                    [
                        'id' => 'marketplace-listings',
                        'module' => 'marketplace',
                        'controller' => 'admin-listings',
                        'params' => [
                            'user_id' => $event->getUserId(),
                        ],
                        'label' => ___('Listings'),
                        'resource' => self::ADMIN_PERM
                    ],
                    [
                        'id' => 'marketplace-transactions',
                        'module' => 'marketplace',
                        'controller' => 'admin-transactions',
                        'params' => [
                            'user_id' => $event->getUserId(),
                        ],
                        'label' => ___('Transactions'),
                        'resource' => self::ADMIN_PERM
                    ],
                ]
            ]);
        }
    }
    
    function onUserMenuItems(Am_Event $event)
    {
        $event->addReturn(function (Am_Navigation_Container $nav, /* ?User */ $user, $order, $config)
        {
            if ($user && $this->isUserHasMarketplaceAccess($user)) {
                return $nav->addPage(
                    [
                        'id' => 'marketplace',
                        'controller' => 'index',
                        'module' => 'marketplace',
                        'label' => ___('Marketplace'),
                        'order' => $order,
                        'visible' => $this->isUserHasMarketplaceAccess($user),
                    ]
                    , true);
            }
        }, 'marketplace');
    }
    
    function isUserHasMarketplaceAccess(User $user): bool
    {
        if($this->getConfig('free-access'))
            return true;
        foreach ($user->getActiveProducts() as $product) {
            if ($product->data()->get(self::PRODUCT_ACCESS)) {
                return true;
            }
        }
        return false;
    }
    
    function isMarketplaceEnabled(User $user): bool
    {
        foreach ($user->getActiveProducts() as $product) {
            if ($product->data()->get(self::PRODUCT_ENABLED)) {
                return true;
            }
        }
        return false;
    }
    
    function isMarketplaceQtyEnabled(User $user): bool
    {
        foreach ($user->getActiveProducts() as $product) {
            if ($product->data()->get(self::PRODUCT_ENABLE_QTY)) {
                return true;
            }
        }
        return false;
    }
    
    function hasOrders(User $user): bool
    {
        return $this->transactionTable()->countBy(['buyer_id' => $user->pk()]);
    }
    
    function hasSales(User $user): bool
    {
        return $this->transactionTable()->countBy(['seller_id' => $user->pk()]);
    }
    
    
    function setupHooks()
    {
        parent::setupHooks(); // TODO: Change the autogenerated stub
    }
    
    /**
     * @return MarketplaceListingTable $table;
     */
    function listingsTable()
    {
        return $this->getDi()->marketplaceListingTable;
    }
    
    /**
     * @return MarketplaceTransactionTable
     */
    function transactionTable()
    {
        return $this->getDi()->marketplaceTransactionTable;
    }
    
    /**
     * @return SelfserviceHistoryTable $table;
     */
    function creditsHistoryTable()
    {
        return $this->getDi()->selfserviceHistoryTable;
    }
    
    /**
     * @return MarketplaceCategoryTable
     */
    function categoryTable()
    {
        return $this->getDi()->marketplaceCategoryTable;
    }
    
    /**
     * @return MarketplaceCategoryTypeTable
     */
    function categoryTypeTable()
    {
        return $this->getDi()->marketplaceCategoryTypeTable;
    }
    
    /**
     * @return MarketplaceTransactionLogTable $table
     */
    function logTable()
    {
        return $this->getDi()->marketplaceTransactionLogTable;
    }
    
    
    function onDaily()
    {
        $this->updateRatings();
    }
    
    
    function updateRatings()
    {
        foreach ($this->getDi()->db->selectCol("select distinct seller_id from ?_marketplace_transaction") as $seller_id) {
            $user = $this->getDi()->userTable->load($seller_id);
            $user->marketplace_rating = $this->transactionTable()->getSellerRating($seller_id);
            $user->marketplace_transactions = $this->transactionTable()->getTotalSold($seller_id);
            $user->update();
        }
    }
    
    function onSetupEmailTemplateTypes(Am_Event $e)
    {
        $e->addReturn([
            'id' => 'marketplace.admin_new_listing',
            'title' => 'New Listing Notification to Admin',
            'mailPeriodic' => Am_Mail::REGULAR,
            'isAdmin' => true,
            'vars' => ['user', 'listing', 'url']
        ], 'marketplace.admin_new_listing');
    
        $e->addReturn([
            'id' => 'marketplace.new_sale',
            'title' => 'New Sale Notification to seller',
            'mailPeriodic' => Am_Mail::REGULAR,
            'isAdmin' => false,
            'vars' => ['transaction', 'url', 'status', 'user']
        ], 'marketplace.new_sale');
        $e->addReturn([
            'id' => 'marketplace.transaction_status_changed_seller',
            'title' => 'Transaction updated notification for seller',
            'mailPeriodic' => Am_Mail::REGULAR,
            'isAdmin' => false,
            'vars' => ['transaction', 'url', 'status', 'user']
        ], 'marketplace.transaction_status_changed_seller');
        $e->addReturn([
            'id' => 'marketplace.transaction_status_changed_buyer',
            'title' => 'Transaction updated notification for buyer',
            'mailPeriodic' => Am_Mail::REGULAR,
            'isAdmin' => false,
            'vars' => ['transaction', 'url', 'status', 'user']
        ], 'marketplace.transaction_status_changed_buyer');
    
        $e->addReturn([
            'id' => 'marketplace.deliver_item',
            'title' => 'Request to deliver item',
            'mailPeriodic' => Am_Mail::REGULAR,
            'isAdmin' => false,
            'vars' => ['transaction', 'url', 'hours', 'user']
        ], 'marketplace.deliver_item');
        $e->addReturn([
            'id' => 'marketplace.confirm_sale',
            'title' => 'Request to confirm order',
            'mailPeriodic' => Am_Mail::REGULAR,
            'isAdmin' => false,
            'vars' => ['transaction', 'url', 'hours', 'status', 'seller', 'user']
        ], 'marketplace.confirm_sale');
        
    }
    
    function newListingNotificationToAdmin(MarketplaceListing $listing)
    {
        if ($this->getConfig('admin_new_listing')) {
            if ($et = Am_Mail_Template::load('marketplace.admin_new_listing')) {
                $et->setListing($listing)
                    ->setUser($listing->getUser())
                    ->setUrl($this->getDi()->surl("marketplace/admin-listings-na?_admin_listings_na_a=edit&_admin_listings_na_b=%2Famember%2Fmarketplace%2Fadmin-listings-na&_admin_listings_na_id={$listing->pk()}"))
                    ->sendAdmin();
            }
        }
    }
    
    function newSaleNotification(MarketplaceTransaction $transaction)
    {
        if ($et = Am_Mail_Template::load('marketplace.new_sale')) {
            $et->setTransaction($transaction)
                ->setUrl($this->getDi()->surl("marketplace/history/sales/{$transaction->pk()}"))
                ->setStatus(MarketplaceTransaction::statusText($transaction->status))
                ->setUser($transaction->getSeller())
                ->send($transaction->getSeller());
        }
        
    }
    function transactionStatusChangedNotification(MarketplaceTransaction $transaction)
    {
        if ($et = Am_Mail_Template::load('marketplace.transaction_status_changed_seller')) {
            $et->setTransaction($transaction)
                ->setUrl($this->getDi()->surl("marketplace/history/sales/{$transaction->pk()}"))
                ->setStatus(MarketplaceTransaction::statusText($transaction->status))
                ->setUser($transaction->getSeller())
                ->send($transaction->getSeller());
        }
    }
    function transactionStatusChangedNotificationBuyer(MarketplaceTransaction $transaction)
    {
        if ($et = Am_Mail_Template::load('marketplace.transaction_status_changed_buyer')) {
            $et->setTransaction($transaction)
                ->setUrl($this->getDi()->surl("marketplace/history/orders/{$transaction->pk()}"))
                ->setStatus(MarketplaceTransaction::statusText($transaction->status))
                ->setUser($transaction->getBuyer())
                ->send($transaction->getBuyer());
        }
    }
    
    function deliverItemNotification(MarketplaceTransaction $transaction)
    {
        if ($et = Am_Mail_Template::load('marketplace.transaction_status_changed_seller')) {
            $et->setTransaction($transaction)
                ->setUrl($this->getDi()->surl("marketplace/history/sales/{$transaction->pk()}"))
                ->setStatus(MarketplaceTransaction::statusText($transaction->status))
                ->setUser($transaction->getSeller())
                ->send($transaction->getSeller());
        }
    }
    
    function onHourly()
    {
        $sendTimeframe = $this->getConfig('seller-delivery-timeframe', 36);
        $completeTimeframe = $this->getConfig('buyer-confirm-timeframe', 36);
        
        foreach($this->transactionTable()->selectObjects(
            "select * from ?_marketplace_transactions where status=? and expire<?",
            MarketplaceTransaction::STATUS_PENDING, $this->getDi()->sqlDateTime) as $record)
        {
            /**
             * @var MarketplaceTransaction $record
             */
            $record->revert();
            
            $this->transactionStatusChangedNotification($record);
            $this->transactionStatusChangedNotificationBuyer($record);
        }
        
        foreach($this->transactionTable()->selectObjects(
            "select * from ?_marketplace_transactions where status=? and updated<?",
            MarketplaceTransaction::STATUS_SENT, $this->getDi()->dateTime->modify("-{$completeTimeframe} hours")->format("Y-m-d H:i:s")) as $record)
        {
            /**
             * @var MarketplaceTransaction $record
             */
            $record->status=MarketplaceTransaction::STATUS_EXPIRED;
            $record->update();
            
            $this->transactionStatusChangedNotification($record);
            $this->transactionStatusChangedNotificationBuyer($record);
        }
        
    }
    function onAfterRender(Am_Event_AfterRender $event)
    {
        if(strpos($event->getTemplateName(), 'marketplace/index.phtml')!== false){
            $event->replace('/am-body-content[^-]/', '');
            
        }
    }
    
    function onDbUpgrade()
    {
        print "Marketplace: Update categories...";
        $cnt = $this->getDi()->db->selectCell("select count(*) from ?_marketplace_category where type_id is null");
        if($cnt)
        {
            /**
             * @var MarketplaceCategoryType $record
             */
            $record = $this->categoryTypeTable()->createRecord();
            $record->title = ___('Default');
            $record->slug = 'default';
            $record->description = ___('Default Category');
            $record->sort=1;
            $record->insert();
            $this->getDi()->db->query("update ?_marketplace_category set type_id = ? where type_id is null", $record->pk());
            $this->getDi()->db->query("update ?_marketplace_category_path set type_id = ? where type_id is null", $record->pk());
            $this->getDi()->db->query("update ?_marketplace_category_sort set type_id = ? where type_id is null", $record->pk());
        }
        print "Done";
    }
    
    function onThanksPage(Am_Event $e)
    {
        $data = $this->getPurchaseData();
        if(!empty($data['listing_id']))
        {
            $this->setPurchaseData(null);
            $url = $data['selfUrl']??ROOT_SURL;
            Am_Mvc_Response::redirectLocation($url . '/item/'.$data['listing_id']);
        }
    }
    
    
}