import React, {FC, useState} from "react";
import {
    Box,
    Button,
    Dialog,
    DialogActions,
    DialogContent,
    DialogContentText,
    DialogTitle,
    Grid,
    Typography,
    useTheme
} from "@mui/material";
import DeleteIcon from '@mui/icons-material/Delete';
import EditIcon from '@mui/icons-material/Edit';
import Status from "../Status";
import {Link, useNavigate} from "react-router-dom";
import {ListingRecord} from "../../Store/Listing";
import CategoryList from "./CategoryList";
import OwnerInfo from "./OwnerInfo";
import ListingImage from "./Image";
import RemoveRedEyeIcon from '@mui/icons-material/RemoveRedEye';

export type TListingBoxProps = {
    record: ListingRecord
    showCategory: boolean
    showOwner: boolean
    showOrder: boolean
    showStatus: boolean
    canEdit: boolean
    canDelete: boolean
    onDelete: (id: string) => void
}


const ListingBox: FC<TListingBoxProps> = (props) => {

    const theme = useTheme();
    const [showDeleteDialog, toggleDeleteDialog] = useState(false);
    const handleClose = () => toggleDeleteDialog(false);
    const handleConfirm = () => {
        props.onDelete(props.record.id);
        toggleDeleteDialog(false);
    }

    const navigate = useNavigate();
    const goToOrder = () => props.showOrder ? navigate(`/item/${props.record.id}`) : (props.canEdit ? navigate(`/my/edit/${props.record.id}`) : "");

    return <Grid container
                 flexDirection={"row"}
                 flexWrap={"nowrap"} sx={{
        width: '100%',
        paddingBottom: '0.5em',
        marginTop: '0.5em',
    }}>

        <Grid item container xs={2} sx={{
            alignItems: "center",
            justifyContent: "center",
            minHeight: '100px',
            minWidth: '100px',
            border: `1px solid ${theme.palette.secondary.light}`
        }}>
            <Box onClick={goToOrder} sx={{cursor: "pointer"}}><ListingImage src={props.record.image} maxHeight={"100px"}
                                                                            maxWidth={"100px"}/></Box>
        </Grid>
        <Grid item container xs={6} sx={{padding: "0 0.5em 0"}} alignItems={"flex-start"} justifyContent={"flex-start"}
              flexDirection={"row"}>
            <Typography onClick={goToOrder} variant={"body1"} sx={{cursor: "pointer"}}>{props.record.title}</Typography>
            {props.showCategory && <Grid item xs={12} sx={{padding: "0.5em 0 0"}}>
                <CategoryList categories={props.record.categories}/>
            </Grid>}
            <Grid item xs={12}><Typography variant={"body2"} color={"secondary"}>Item
                Type: {props.record.listingTypeText}</Typography></Grid>
            <Grid item container xs={12} flexDirection={"row"} justifyContent={"space-between"}>
                <Grid item xs={6}><Typography variant={"body2"} color={"secondary"}>Delivery
                    Speed: {props.record.deliveryText}</Typography></Grid>
            </Grid>

            <Grid item container xs={12} sx={{padding: "0.5em 0 0"}} flexDirection={"row"}
                  justifyContent={"space-between"}>
                <Grid item xs={6}><Typography variant={"body2"}
                                              color={"secondary"}>Available: {props.record.qty}</Typography></Grid>
                <Grid xs={6} item container alignItems={"center"} justifyContent={"flex-end"}>
                    <RemoveRedEyeIcon sx={{fontSize: "18px"}} color={"secondary"}/>&nbsp;<Typography variant={"body2"}
                                                                                                     color={"secondary"}>{props.record.views}</Typography>
                </Grid>

            </Grid>

        </Grid>
        {props.showOwner && <Grid item xs={2} sx={{padding: "0 0.5em 0"}}>
            <OwnerInfo record={props.record}/>
        </Grid>}


        <Grid item xs={2} sx={{padding: "0 0.5em 0"}} container justifyContent={"space-between"}
              alignItems={"center"} flexDirection={"column"} alignSelf={"flex-start"} justifySelf={"flex-end"}>
            <Typography variant={"h5"} component={"span"}>{props.record.price}</Typography>

            {props.showOrder &&
                <Button onClick={goToOrder}>Purchase</Button>
            }

        </Grid>

        {(props.showStatus || props.canDelete || props.canEdit) &&
            <Grid item xs={2} sx={{padding: "0 0.5em 0"}} container justifyContent={"space-between"}>
                {props.showStatus && <Grid item container xs flexDirection={"column"}>
                    <Status status={props.record.status}></Status>
                    <Typography component={"span"} variant={'body2'}
                                color={props.record.approved ? theme.palette.success.main : theme.palette.error.main}>{props.record.approved ? "Approved" : "Not Approved"}</Typography>
                </Grid>}

                <Grid item container xs={3} justifyContent={"flex-end"} spacing={2}>
                    {props.canEdit &&
                        <Grid item xs={6}><Link to={`/my/edit/${props.record.id}`}><EditIcon fontSize={"small"}
                                                                                             color={"primary"}/></Link></Grid>
                    }
                    {props.canDelete &&
                        <Grid item xs={6}>
                            <Link to={"/my/delete"} onClick={(event) => {
                                event.preventDefault();
                                toggleDeleteDialog(true)
                            }}><DeleteIcon fontSize={"small"} color={"primary"}/></Link>
                            <Dialog
                                open={showDeleteDialog}
                                onClose={handleClose}
                                aria-labelledby="alert-dialog-title"
                                aria-describedby="alert-dialog-description"
                            >
                                <DialogTitle id="alert-dialog-title">
                                    {"Listing Delete Confirmation"}
                                </DialogTitle>
                                <DialogContent>
                                    <DialogContentText id="alert-dialog-description">
                                        Are you sure you want to delete listing "{props.record.title}"?
                                        This action can't be reverted.
                                    </DialogContentText>
                                </DialogContent>
                                <DialogActions>
                                    <Button onClick={handleClose}>Cancel</Button>
                                    <Button onClick={handleConfirm} autoFocus>
                                        DELETE
                                    </Button>
                                </DialogActions>
                            </Dialog>
                        </Grid>
                    }

                </Grid>

            </Grid>
        }


    </Grid>
}
export default ListingBox;
