import UserStore from "./User";
import {createContext, useContext} from "react";
import {TMarketplaceStoreConfig, TModuleConfig} from "../types";
import {TApiResponse, TApiSuccessResponse} from "../../types";
import {makeAutoObservable, runInAction} from "mobx";
import postData from "../../Utils/postData";
import ListingStore from "./Listing";
import TransactionsStore from "./Transactions";
import CategoryTypes from "./CategoryTypes";

export enum TStoreStatus {
    IDLE,
    LOADING
}

export class MarketplaceStore {
    userStore: UserStore
    listingStore: ListingStore
    private amUrl: string
    selfUrl: string
    isError: boolean = false
    errorMessage: string = ""
    status: TStoreStatus = TStoreStatus.LOADING
    config: TModuleConfig;
    transactions: TransactionsStore
    categories: CategoryTypes;

    constructor(config: TMarketplaceStoreConfig) {
        this.amUrl = config.amUrl
        this.selfUrl = config.selfUrl
        this.userStore = new UserStore(this)
        this.listingStore = new ListingStore(this)
        this.transactions = new TransactionsStore(this)
        this.categories = new CategoryTypes(this);

        this.config = {
            maxUploadWidth: 200,
            maxUploadSize: 200,
            buyCreditsUrl: ""

        }
        makeAutoObservable(this, {
            userStore: false,
            listingStore: false

        });
    }

    async init(): Promise<any> {
        runInAction(() => {
            this.status = TStoreStatus.LOADING
        })

        let loadConfig = this.loadConfig();

        let fetchUser = this.userStore.fetchUser();
        let fetchCategories = this.categories.apiLoadTypes();

        await Promise.all([loadConfig, fetchUser, fetchCategories]).then(() => {
            runInAction(() => {
                this.status = TStoreStatus.IDLE;
            })
        })

    }

    apiUrl(endpoint: string): string {
        return this.amUrl + '/marketplace/ajax' + endpoint;
    }

    async loadConfig() {
        const response = await this.apiRequest("/config").then((response) => {
            this.config = Object.assign({}, this.config, (response as TApiSuccessResponse).config);
        });
        return response;
    }

    async apiRequest(endpoint: string, data: {} = {}): Promise<TApiResponse> {

        try {
            return await postData(this.apiUrl(endpoint), data);
        } catch (error) {
            runInAction(() => {
                this.setError((error as Error).message);
            })
            return {ok: false, errorMessage: (error as Error).message, errorCode: 500}

        }
    }

    async apiUpload(file: File): Promise<TApiResponse> {

        try {
            const data = new FormData()
            data.append('upload', file);

            const response = await fetch(this.apiUrl('/upload'), {
                method: 'POST',
                mode: 'cors',
                cache: 'no-cache',
                credentials: 'same-origin',
                redirect: 'follow',
                referrerPolicy: 'no-referrer',
                body: data
            });

            const json = await response.json();

            if (json.ok === false)
                throw new Error(json.errorMessage);

            return json;

        } catch (error) {
            runInAction(() => {
                this.setError((error as Error).message);
            })
            return {ok: false, errorMessage: (error as Error).message, errorCode: 500}
        }
    }


    setError(message: string) {
        this.isError = true;
        this.errorMessage = message;
    }

    clearError() {
        this.isError = false;
        this.errorMessage = "";
    }

    getConfigValue(key: keyof TModuleConfig) {
        return this.config[key] ?? undefined;
    }

    get loginUrl() {
        const u = this.amUrl + '/login';
        const search = new URLSearchParams();
        search.append('_amember_redirect_url', btoa(document.location.href));
        return u + '?' + search.toString();
    }

}

export const MarketplaceStoreContext = createContext<MarketplaceStore>({} as MarketplaceStore);
export const useMarketplaceStore = () => useContext(MarketplaceStoreContext);
export const useListingStore = () => {
    const rootStore = useMarketplaceStore();
    return rootStore.listingStore;
}

export const useCurrentUser = () => {
    const rootStore = useMarketplaceStore();
    return rootStore.userStore.user;
}