import React, {useEffect, useState} from "react";
import {observer} from "mobx-react";
import {useCategoryStore} from "../../AdminCategory/Store";
import {TListingFormError, TListingFormValue} from "../types";
import {useMarketplaceStore} from "../Store";
import {useNavigate, useParams} from "react-router-dom";
import {Grid} from "@mui/material";
import PageHeader from "../Layout/PageHeader";
import Loading from "../../Utils/Loading";
import MarketplaceLayout from "../Layout";
import {TApiSuccessResponse} from "../../types";

const ListingForm = React.lazy(() => import("../Layout/Listing/Form"));


const EditListing = observer(() => {
    const categoryStore = useCategoryStore();
    const [formError, setFormError] = useState<TListingFormError | undefined>(undefined);
    const store = useMarketplaceStore();
    const navigate = useNavigate();
    const [isFormLoading, toggleFormLoading] = useState(false);
    const {id} = useParams();
    const [listing, setListing] = useState<TListingFormValue>({} as TListingFormValue);
    const [dataLoaded, toggleDataLoaded] = useState(false);

    if (!id)
        throw Error("ID is empty");

    useEffect(() => {
        toggleDataLoaded(false);

        store.apiRequest('/listing/get', {listing_id: id}).then((response) => {
            if ('status' in response) {
                if (response.status == 'error') {
                    setFormError(response.error);
                } else {
                    if (typeof response.listing === 'object' && 'listing_id' in response.listing)
                        setListing(response.listing as unknown as TListingFormValue);
                    else
                        setFormError("Unable to find listing")
                }
            }

        }).finally(() => {
            toggleDataLoaded(true);
        })
    }, [id])

    const onSubmit = (formValue: TListingFormValue) => {
        toggleFormLoading(true);
        const upload = formValue.upload ? store.apiUpload(formValue.upload) : Promise.resolve({
            ok: true,
            status: 'ok',
            error: ""
        } as unknown as TApiSuccessResponse);

        upload.then((response) => {
            if ('upload_id' in response)
                formValue.image = response.upload_id as string;

            return response.ok ? store.apiRequest('/listing/save', {form: formValue, listing_id: id}) : response

        }).then((response) => {
            if ('status' in response) {
                if (response.status == 'error') {
                    return Promise.reject(response.error);
                }
            }
            navigate('/my');
        }).catch((error) => {
            setFormError(error);
        }).finally(() => {
            toggleFormLoading(false);
        })

    }


    return (
        <MarketplaceLayout>
            {store.userStore.user && <Grid container spacing={2}>
                <Grid item xs={12}>
                    <PageHeader>Edit Listing</PageHeader>
                </Grid>
                <Grid item xs={12} lg={8}>
                    {dataLoaded ? <ListingForm
                        listing_id={parseInt(id)}
                        listing={listing}
                        categoryOptions={categoryStore.categorySelectOptions}
                        user={store.userStore.user}
                        onCancel={() => navigate('/my', {replace: true})}
                        onSubmit={onSubmit}
                        isLoading={isFormLoading}
                        error={formError}
                        isDisabled={!['draft', 'published'].includes(listing.status)}
                    /> : <Loading/>}

                </Grid>

            </Grid>
            }

        </MarketplaceLayout>

    )
})
export default EditListing;